import functools
from collections import defaultdict


import connexion
from flask import abort  # TODO: change to connextion abort
import sqlalchemy
import marshmallow

from paddlelabel.config import db
from paddlelabel.api.util import parse_order_by

# TODO: implement a search method
def crud(Model, Schema, triggers=[]):
    tgs = defaultdict(lambda: None)
    for trigger in triggers:
        tgs[trigger.__name__] = trigger

    def get_all(
        Model,
        Schema,
        order_by="modified desc",
        pre_get_all=tgs["pre_get_all"],
        post_get_all=tgs["post_get_all"],
    ):
        order = parse_order_by(Model, order_by)

        items = Model.query.order_by(order).all()
        # print(items)
        if post_get_all is not None:
            post_get_all(items, db.session)
        return Schema(many=True).dump(items), 200

    def get(Model, Schema, **kwargs):
        id_name, id_val = list(kwargs.items())[0]
        item = Model.query.filter(getattr(Model, id_name) == id_val).one_or_none()

        if item is not None:
            return Schema().dump(item), 200
        abort(404, f"No {id_name.split('_')[0]} with id: {id_val}")

    def post(
        Model,
        Schema,
        pre_add=tgs["pre_add"],
        post_add=tgs["post_add"],
    ):
        schema = Schema()
        try:
            new_item = schema.load(connexion.request.json)
        except marshmallow.exceptions.ValidationError as e:
            for field, msgs in e.messages.items():
                if "Missing data for required field." in msgs:
                    # TODO: change code
                    abort(
                        500,
                        f"Marshmallow catch: Missing data for required field: {field}",
                    )
            abort(500, e.messages)
        if pre_add is not None:
            new_item = pre_add(new_item, db.session)
        try:
            db.session.add(new_item)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError as e:
            msg = str(e.orig)
            if msg.startswith("UNIQUE constraint failed"):
                col = msg.split(":")[1].strip()
                abort(
                    409,
                    f"{Model.__tablename__} doesn't allow duplicate {col}.",
                )
            else:
                abort(500, msg)

        if post_add is not None:
            with db.session.no_autoflush:
                post_add(new_item, db.session)
        return schema.dump(new_item), 201

    def put(
        Model,
        Schema,
        pre_put=tgs["pre_put"],
        post_put=tgs["post_put"],
        **kwargs,
    ):
        # 1. check item exist
        id_name, id_val = list(kwargs.items())[0]
        item = Model.query.filter(getattr(Model, id_name) == id_val).one_or_none()
        if item is None:
            abort(
                404,
                f"No {Model.__tablename__} with {id_name}: {id_val} .",
            )
        # 2. check request key exist and can be edited
        body = connexion.request.json

        for k in list(body.keys()):
            if k in Model._immutables:
                # abort(403, f"{Model.__tablename__}.{k} doesn't allow edit")
                del body[k]
            if k not in Model._cols:
                abort(404, f"{Model.__tablename__}.{k} doesn't have property {k}")

        if pre_put is not None:
            item, body = pre_put(item, body, db.session)

        # 3. edit item
        Model.query.filter(getattr(Model, id_name) == id_val).update(body)
        db.session.commit()

        # FIXME: really need to requery?
        item = Model.query.filter(getattr(Model, id_name) == id_val).one()
        if post_put is not None:
            post_put(item, db.session)
        return Schema().dump(item), 200

    def delete(
        Model,
        Schema,
        pre_delete=tgs["pre_delete"],
        post_delete=tgs["post_delete"],
        **kwargs,
    ):
        id_name, id_val = list(kwargs.items())[0]
        item = Model.query.filter(getattr(Model, id_name) == id_val).one_or_none()

        if item is None:
            # abort(404, f"No {Model.__tablename__} with {id_name} == {id_val}")
            return

        if pre_delete is not None:
            item = pre_delete(item, db.session)
        db.session.delete(item)

        if post_delete is not None:
            post_delete(item, db.session)

        db.session.commit()

        return f"{Model.__tablename__.capitalize()} {id_val} deleted", 200

    get_all = functools.partial(get_all, Model, Schema)
    get = functools.partial(get, Model, Schema)
    post = functools.partial(post, Model, Schema)
    put = functools.partial(put, Model, Schema)
    delete = functools.partial(delete, Model, Schema)
    return get_all, get, post, put, delete



import connexion
from paddlelabel.config import db
from .base import crud
from ..model import Task, Project
from ..schema import TaskSchema
from paddlelabel.api.util import abort, parse_order_by

# TODO: reject tasks with same datas
get_all, get, post, put, delete = crud(Task, TaskSchema)


def get_by_project(project_id, order_by="created asc"):
    if connexion.request.method == "HEAD":
        return get_stat_by_project(project_id)
    Project._exists(project_id)
    order = parse_order_by(Task, order_by)

    tasks = Task.query.filter(Task.project_id == project_id).order_by(order).all()
    return TaskSchema(many=True).dump(tasks), 200


# TODO: dont lazy load annotations in tasks
def get_stat_by_project(project_id):
    Project._exists(project_id)
    tasks = Task.query.filter(Task.project_id == project_id).all()
    ann_count = 0
    for task in tasks:
        if len(task.annotations) != 0:
            ann_count += 1
    res = {"finished": ann_count, "total": len(tasks)}
    return res, 200, res



import os.path as osp
import shutil

from paddlelabel.config import db
from paddlelabel.api import Task
from paddlelabel.task.util.file import copycontent
from .util import create_dir, listdir, copy, image_extensions
from .base import BaseTask


class Classification(BaseTask):
    def __init__(self, project, data_dir=None):
        super(Classification, self).__init__(project, data_dir=data_dir)
        self.importers = {
            "single_class": self.single_class_importer,
            "multi_class": self.multi_class_importer,
        }
        self.exporters = {
            "single_class": self.single_class_exporter,
            "multi_class": self.multi_class_exporter,
        }
        self.default_importer = self.single_class_importer
        self.default_exporter = self.single_class_exporter

    def single_class_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        # 1. set params
        project = self.project
        if data_dir is None:
            data_dir = project.data_dir

        self.create_warning(data_dir)
        # 2. import all datas,
        data_paths = listdir(data_dir, filters)
        for data_path in data_paths:
            label_name = osp.basename(osp.dirname(data_path))
            label = [{"label_name": label_name}] if len(label_name) != 0 else []
            self.add_task([{"path": data_path}], [label])

        db.session.commit()

    def multi_class_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        # 1. set params
        project = self.project
        if data_dir is None:
            data_dir = project.data_dir
        self.create_warning(data_dir)

        other_settings = self.project._get_other_settings()
        print("other_settings", other_settings)
        if other_settings is not None:
            delimiter = other_settings.get("xx_list_delimiter", " ")
        else:
            delimiter = " "

        # 2. get label names from xx_list.txt
        label_lines = []
        for list_name in ["train_list.txt", "val_list.txt", "test_list.txt"]:
            list_path = osp.join(data_dir, list_name)
            if osp.exists(list_path):
                label_lines += open(list_path, "r").readlines()
        label_lines = [l.strip().split(delimiter) for l in label_lines if l.strip() != ""]

        labels_dict = {}
        for l in label_lines:
            labs = []
            for lab in l[1:]:
                try:
                    # if number, get name
                    lab_idx = int(lab) + 1
                    labs.append(self.label_id2name(lab_idx))
                except ValueError:
                    # if str, use str as name
                    labs.append(lab)

            labels_dict[osp.normpath(l[0])] = labs

        data_paths = listdir(data_dir, filters)
        for data_path in data_paths:
            labels = labels_dict.get(osp.normpath(data_path), [])
            self.add_task([{"path": data_path}], [[{"label_name": name} for name in labels]])

        db.session.commit()

    def single_class_exporter(self, export_dir):
        project = self.project
        create_dir(export_dir)
        create_dir(osp.join(export_dir, "no_annotation"))
        have_no_annotation = False

        # 1. write labels.txt
        labels = self.export_labels(export_dir)

        # 2. create label dirs
        for label in labels:
            create_dir(osp.join(export_dir, label.name))

        # 3. move files to output dir
        tasks = Task._get(project_id=project.project_id, many=True)
        new_paths = []
        for task in tasks:
            for data in task.datas:
                label_name = ""
                if len(data.annotations) == 0:
                    label_name = "no_annotation"
                    have_no_annotation = True
                else:
                    label_name = data.annotations[0].label.name
                copy(osp.join(project.data_dir, data.path), osp.join(export_dir, label_name))
                new_paths.append([osp.join(label_name, osp.basename(data.path))])

        if not have_no_annotation:
            shutil.rmtree(osp.join(export_dir, "no_annotation"))

        # 4. write split files
        self.export_split(export_dir, tasks, new_paths)

    def multi_class_exporter(self, export_dir):
        project = self.project

        # 1. all images
        # with annotation go to export_dir/image
        # without annotation go to export_dir/no_annotation
        create_dir(osp.join(export_dir, "image"))
        create_dir(osp.join(export_dir, "no_annotation"))
        have_no_annotation = False

        # 2. write labels.txt
        self.export_labels(export_dir)

        # 3. move all images to image folder
        tasks = Task._get(project_id=project.project_id, many=True)
        new_paths = []
        for task in tasks:
            for data in task.datas:
                if len(data.annotations) == 0:
                    folder = "no_annotation"
                    have_no_annotation = True
                else:
                    folder = "image"

                copy(
                    osp.join(project.data_dir, data.path),
                    osp.join(export_dir, folder, osp.basename(data.path)),
                )
                new_paths.append([osp.join(folder, osp.basename(data.path))])
        if not have_no_annotation:
            shutil.rmtree(osp.join(export_dir, "no_annotation"))

        # 4. export split
        self.export_split(osp.join(export_dir), tasks, new_paths)



import string
import random

import connexion


def rand_string(length):
    chars = string.ascii_letters + string.punctuation
    return "".join(random.choice(chars) for x in range(20))


def camel2snake(string):
    """
    convert camel case to snake case
    """
    if string is None:
        return None
    new_string = ""
    for c in string:
        if c >= "A" and c <= "Z":
            new_string += "_" + c.lower()
        else:
            new_string += c
    return new_string


class Resolver(connexion.resolver.RestyResolver):
    def resolve_operation_id_using_rest_semantics(self, operation):
        """
        Resolves the operationId using REST semantics
        :type operation: connexion.operations.AbstractOperation
        """

        # Split the path into components delimited by '/'
        path_components = [c for c in operation.path.split("/") if len(c)]

        def is_var(component):
            """True if the path component is a var. eg, '{id}'"""
            return (component[0] == "{") and (component[-1] == "}")

        resource_name = ".".join([c for c in path_components if not is_var(c)]).replace("-", "_")
        if resource_name[-1] == "s":
            resource_name = resource_name[:-1]

        def get_controller_name():
            x_router_controller = operation.router_controller

            name = self.default_module_name

            if x_router_controller:
                name = x_router_controller

            elif resource_name:
                d = self.default_module_name
                r = resource_name
                # name = f"{d}.{r}.controller"
                name = f"{d}.controller.{r}"
            return name

        def get_function_name():
            method = operation.method

            is_collection_endpoint = method.lower() == "get" and len(resource_name) and not is_var(path_components[-1])

            return self.collection_endpoint_name if is_collection_endpoint else method.lower()

        # print(f"{get_controller_name()}.{get_function_name()}")
        return f"{get_controller_name()}.{get_function_name()}"

    # TODO: find a better way to resolve this
    def resolve_operation_id(self, operation):
        # TODO: auto resolve /collection/{item}/collection

        special = {
            "/projects/{project_id}/tasks getTasks": "paddlelabel.api.controller.task.get_by_project",
            "/projects/{project_id}/labels getLabels": "paddlelabel.api.controller.label.get_by_project",
            "/projects/{project_id}/labels setLabels": "paddlelabel.api.controller.label.set_by_project",
            "/projects/{project_id}/labels removeLabels": "paddlelabel.api.controller.label.delete_by_project",
            "/projects/{project_id}/annotations getAnnotations": "paddlelabel.api.controller.annotation.get_by_project",
            "/projects/{project_id}/tags getTags": "paddlelabel.api.controller.tag.get_by_project",
            "/projects/{project_id}/progress getProgress": "paddlelabel.api.controller.task.get_stat_by_project",
            "/projects/{project_id}/split splitDataset": "paddlelabel.api.controller.project.split_dataset",
            "/projects/{project_id}/export exportDataset": "paddlelabel.api.controller.project.export_dataset",
            "/projects/{project_id}/import importDataset": "paddlelabel.api.controller.project.import_dataset",
            "/projects/{project_id}/predict predict": "paddlelabel.api.controller.project.predict",
            "/tasks/{task_id}/tags getTags": "paddlelabel.api.controller.tag.get_by_task",
            "/tasks/{task_id}/tags addTag": "paddlelabel.api.controller.tag.add_to_task",
            "/datas/{data_id}/image getImage": "paddlelabel.api.controller.data.get_image",
            "/tasks/{task_id}/annotations getAnnotations": "paddlelabel.api.controller.annotation.get_by_task",
            "/tasks/{task_id}/datas getDatas": "paddlelabel.api.controller.data.get_by_task",
            "/datas/{data_id}/annotations getAnnotations": "paddlelabel.api.controller.annotation.get_by_data",
            "/datas/{data_id}/annotations setAnnotations": "paddlelabel.api.controller.annotation.set_all_by_data",
            "/datas/{data_id}/annotations deleteAnnotations": "paddlelabel.api.controller.annotation.delete_by_data",
            "/rpc/folders getFolders": "paddlelabel.api.rpc.file.get_folders",
            "/rpc/seg/polygon2points polygon2points": "paddlelabel.api.rpc.seg.polygon2points_str",
            "/rpc/seg/points2polygon points2polygon": "paddlelabel.api.rpc.seg.points2polygon_str",
            "/version getVersion": "paddlelabel.api.rpc.monitor.get_version",
            "/samples loadSample": "paddlelabel.api.controller.sample.load_sample",
            "/samples/structure getStructure": "paddlelabel.api.controller.sample.sample_folder_structure",
            "/samples/file getFile": "paddlelabel.api.controller.sample.serve_sample_file",
            "/datas/{data_id}/mask getMask": "paddlelabel.api.controller.data.get_mask",
        }
        opid = None

        if operation.operation_id and operation.operation_id.startswith("paddlelabel"):
            opid = operation.operation_id

        path = operation.path
        if path[:-1] == "/":
            path = path[:-1]
        idx = f"{path} {operation.operation_id}"
        if idx in special.keys():
            opid = special[idx]
        # print("resolve resolve resolve", operation.operation_id, path, opid)
        if opid:
            router_controller = operation.router_controller
            if router_controller is None:
                return opid
            return f"{router_controller}.{opid}"

        return self.resolve_operation_id_using_rest_semantics(operation)



import os
import os.path as osp
import json
from collections import deque

import cv2

from paddlelabel.api import Annotation, Data, Label, Project, Task
from paddlelabel.api.util import abort
from paddlelabel.config import db
from paddlelabel.task.util import image_extensions, listdir, create_dir
from paddlelabel.task.util.color import rgb_to_hex, rand_hex_color, name_to_hex

"""
Base for import/export and other task specific operations.
"""


class BaseTask:
    def __init__(self, project, skip_label_import=False, data_dir=None):
        """
        Args:
            project (int|dict): If the project exists, self.project will be queried from db with parameter project as project_id or with project.project_id. Else the project will be created.
        """

        # 1. set project
        if isinstance(project, int):
            curr_project = Project._get(project_id=project)
            if curr_project is None:
                raise RuntimeError(f"No project with project_id {project}")
        else:
            curr_project = Project._get(project_id=project.project_id)
            if curr_project is None:
                db.session.add(project)
                db.session.commit()
                curr_project = project
        self.project = curr_project

        if data_dir is None:
            data_dir = self.project.data_dir

        # os.makedirs(data_dir, exist_ok=True)

        # 2. set current label max id
        # next added label will have id label_max_id+1, so label.id starts from 1
        self.label_max_id = 0
        for label in project.labels:
            self.label_max_id = max(self.label_max_id, label.id)

        # 3. read dataset split
        self.split = self.read_split(data_dir)

        # 4. create labels specified in labels.txt
        if not skip_label_import:
            self.import_labels()

        # 5. polupate label colors
        self.populate_label_colors()

        # 6. get curr datapaths
        # tasks = Task._get(project_id=project.project_id, many=True)
        # self.curr_data_paths = []
        # for task in tasks:
        #     for data in task.datas:
        #         self.curr_data_paths.append(data.path)
        # print("Current data paths", self.curr_data_paths)

        self.project = Project._get(project_id=project.project_id)

    def add_task(self, datas: list, annotations: list = None, split: int = None):
        """Add one task to project.
        ATTENTION: to be more efficient, this method WONT connmit! make sure you invoke db.session.commit() after adding all tasks

        Parameters
        ----------
        datas : list. each item is a dict, path is required, specifying full or relative path to project.data_dir. Others are optional
            [{"path": 'path1'}, {"path" : 'path2', "size": [1024, 768, 3]}, ...]
        annotations : list
            [
                [ // labels for path1
                    {
                        "label_name": "",
                        "result": "", // optional, default to ""
                    },
                    {
                        "label_name": "",
                        "result": "",
                    }
                ],
                [ // labels for path2
                    {
                        "label_name": "",
                        "result": "", // optional, default to ""
                    },
                    {
                        "label_name": "",
                        "result": "", // optional, default to ""
                    }
                ],
                ...
            ]
        split: int. the split set this task is in. If not passed will attempt to find in the three list files. If not found default to 0 (training set). 0, 1, 2-> train, val, test
        """
        project = self.project
        assert len(datas) != 0, "can't add task without data"
        # print(datas, project.data_dir)

        for idx in range(len(datas)):
            if osp.isabs(datas[idx]["path"]):
                datas[idx]["path"] = osp.relpath(datas[idx]["path"], project.data_dir)

        # 1. find task split
        if split is None:
            split_idx = 0
            for idx, split in enumerate(self.split):
                if datas[0]["path"] in split:
                    split_idx = idx
                    break
        else:
            split_idx = split

        task = Task(project_id=project.project_id, set=split_idx)

        def get_label(name):
            for lab in project.labels:
                if lab.name == name:
                    return lab
            return None

        if annotations is None:
            annotations = []
        while len(annotations) < len(datas):
            annotations.append([])

        for anns, data_record in zip(annotations, datas):
            # 2. add data record
            data = Data(**data_record)
            task.datas.append(data)
            total_anns = 0

            # 3. add data's annotations
            for ann in anns:
                if len(ann.get("label_name", "")) == 0:
                    continue
                # BUG: multiple labels under same label_name can exist
                label = get_label(ann["label_name"])
                if label is None:
                    label = self.add_label(ann["label_name"], ann.get("color"), commit=True)
                del ann["label_name"]
                ann = Annotation(label_id=label.label_id, project_id=project.project_id, **ann)
                task.annotations.append(ann)  # TODO: remove
                data.annotations.append(ann)
                total_anns += 1
            print(f"==== {data_record['path']} with {total_anns} annotation(s) imported to set {split_idx} ====")

        db.session.add(task)

    def label_id2name(self, label_id):
        for label in self.project.labels:
            if label.id == label_id:
                return label.name
        return None

    def label_name2id(self, label_name):
        for label in self.project.labels:
            if label.name == label_name:
                return label.id
        return None

    def label_name2label_id(self, label_name):
        for label in self.project.labels:
            if label.name == label_name:
                return label.label_id
        return None

    def read_split(self, data_dir=None, delimiter=" "):
        if data_dir is None:
            data_dir = self.project.data_dir

        sets = []
        split_names = ["train_list.txt", "val_list.txt", "test_list.txt"]
        for split_name in split_names:
            split_path = osp.join(data_dir, split_name)
            paths = []
            if osp.exists(split_path):
                paths = open(split_path, "r").readlines()
                paths = [p.strip() for p in paths if len(p.strip()) != 0]
                paths = [p.split(delimiter)[0] for p in paths]
            sets.append(set(paths))
        return sets

    def export_split(
        self,
        export_dir,
        tasks,
        new_paths,
        delimiter=" ",
        with_labels=True,
        annotation_ext=None,
    ):
        # only used in file-file split, not in file-class split
        if annotation_ext is not None and annotation_ext[0] == ".":
            annotation_ext = annotation_ext[1:]

        set_names = ["train_list", "val_list", "test_list"]
        create_dir(export_dir)
        set_files = [open(osp.join(export_dir, f"{n}.txt"), "w") for n in set_names]
        for task, task_new_paths in zip(tasks, new_paths):
            for data, new_path in zip(task.datas, task_new_paths):
                if with_labels:
                    label_ids = []
                    for ann in data.annotations:
                        label_ids.append(ann.label.id - 1)
                    if len(label_ids) == 0:
                        continue
                    label_ids = [str(id) for id in label_ids]
                    print(
                        new_path + delimiter + delimiter.join(label_ids),
                        file=set_files[task.set],
                    )
                else:
                    annotation_path = new_path.replace("JPEGImages", "Annotations")
                    annotation_path = annotation_path[: -annotation_path[::-1].find(".")] + annotation_ext
                    print(new_path + delimiter + annotation_path, file=set_files[task.set])

        for f in set_files:
            f.close()

    """ label related """

    def add_label(
        self,
        name: str,
        id: int = None,
        color: str = None,
        super_category_id: int = None,
        comment: str = None,
        commit=False,
    ):
        """
        Add one label to current project

        Args:
            name (str): label name
            id (int, optional): id. Defaults to None, autoincrement.
            color (str, optional): the color this label uses, can be hex color with leading # or name for a common color. will raise runtime error if specified color is in use by other labels. Defaults to None, will randomly generate.
            comment (str, optional): comment for label. Defaults to None.
            super_category_id (int, optional): id of supercategory. Defaults to None.
            commit (bool, optional): True -> commit after adding label. Defaults to False.

        Returns:
            Label: new label generated
        """
        # 1. check params
        if name is None or len(name) == 0:
            raise RuntimeError(f"Label name is required, got {name}")
        current_names = set(l.name for l in self.project.labels)
        if name in current_names:
            # raise RuntimeError(f"Label name {name} is not unique")
            print(f"Label {name} already exist, skipping.")
            return

        # 2. check or assign color
        current_colors = set(l.color for l in self.project.labels)
        if color is None:
            color = rand_hex_color(current_colors)
        else:
            if color[0] != "#":
                color = name_to_hex(color)
            if color in current_colors:
                raise RuntimeError(f"Label color {color} is not unique")

        # 3. check or assign id
        current_ids = set(int(l.id) for l in self.project.labels)
        if id is None:
            id = self.label_max_id + 1
        else:
            id = int(id)
            if id in current_ids:
                raise RuntimeError(f"Label id {id} is not unique")

        # 4. assign super category id

        label = Label(
            project_id=self.project.project_id,
            id=id,
            name=name,
            color=color,
            comment=comment,
            super_category_id=super_category_id,
        )
        self.project.labels.append(label)
        if commit:
            db.session.commit()
        current_ids = [l.id for l in self.project.labels]

        self.label_max_id = max(current_ids)
        return label

    def import_labels(self, delimiter=" ", ignore_first=False):
        # 1. set params
        label_names_path = None
        project = self.project
        if project.other_settings is not None:
            if isinstance(project.other_settings, str):
                other_settings = json.loads(project.other_settings)
            else:
                other_settings = project.other_settings
            label_names_path = other_settings.get("label_names_path", None)

        if label_names_path is None:
            label_names_path = osp.join(project.data_dir, "labels.txt")

        if label_names_path is None or not osp.exists(label_names_path):
            return

        # 2. import labels
        labels = open(label_names_path, "r").readlines()
        labels = [l.strip() for l in labels if len(l.strip()) != 0]
        if ignore_first:
            background_line = labels[0]
            labels = labels[1:]

        labels = [l.split("//") for l in labels]
        comments = [None if len(l) == 1 else l[1].strip() for l in labels]
        labels = [l[0].strip() for l in labels]
        labels = [l.split(delimiter) for l in labels]

        current_labels = Label._get(project_id=self.project.project_id, many=True)
        current_labels = [l.name for l in current_labels]
        for label, comment in zip(labels, comments):
            """
            label length: 1: label name
                          2: label name | label id
                          3: label name | label id | hex color or common color name or grayscale value
                          5: label name | label id | r | g | b color
                          //: string after // is stored as comment
                          -: skip this field
            """
            valid_lengths = [1, 2, 3, 5]
            if len(label) not in valid_lengths:
                raise RuntimeError(f"After split got {label}. It's not in valid lengths {valid_lengths}")
            if label[0] not in current_labels:
                print("==== Adding label", label, "====")
                if len(label) == 5:
                    label[2] = rgb_to_hex(label[2:])
                    del label[3]
                label = [None if v == "-" else v for v in label]
                self.add_label(*label, comment=comment)
        db.session.commit()

        if ignore_first:
            return background_line

    def populate_label_colors(self):
        labels = Label._get(project_id=self.project.project_id, many=True)
        for lab in labels:
            if lab.color is None:
                lab.color = rand_hex_color([l.color for l in labels])
        db.session.commit()

    def export_labels(self, export_dir: str, background_line: str = None, with_id: bool = False):
        label_names_path = osp.join(export_dir, "labels.txt")
        labels = self.project.labels
        labels.sort(key=lambda l: l.id)
        with open(label_names_path, "w") as f:
            if background_line is not None:
                print(background_line.strip(), file=f)
            for lab in labels:
                # print(lab)
                print(lab.name, end=" " if with_id else "\n", file=f)
                if with_id:
                    print(lab.id, file=f)
        return labels

    # TODO: add total imported count
    def default_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
        with_size=False,
    ):
        if data_dir is None:
            data_dir = self.project.data_dir

        for data_path in listdir(data_dir, filters):
            if with_size:
                img = cv2.imread(osp.join(data_dir, data_path))
                size = [1] + list(img.shape)
                size = ",".join([str(s) for s in size])
                self.add_task([{"path": data_path, "size": size}])
            else:
                self.add_task([{"path": data_path}])
        db.session.commit()

    """ warning file """

    def create_warning(self, dir):
        if not osp.exists(dir):
            abort(detail=f"Dataset Path specified {dir} doesn't exist.", status=404)

        warning_path = osp.join(dir, "paddlelabel.warning")
        if not osp.exists(warning_path):
            print(
                "PP Label is using files stored under this folder!\nChanging file in this folder may cause issues.",
                file=open(warning_path, "w"),
            )

    def remove_warning(self, dir):
        warning_path = osp.join(dir, "paddlelabel.warning")
        if osp.exists(warning_path):
            os.remove(warning_path)

    def create_coco_labels(self, labels):
        catgs = deque()

        for catg in labels:
            catgs.append(catg)

        tried_names = []  # guard against invalid dependency graph
        for _ in range(len(catgs) * 2):
            if len(catgs) == 0:
                break
            catg = catgs.popleft()
            if self.label_name2id(catg["name"]) is not None:
                continue

            color = catg.get("color", None)
            if color is not None:
                color = name_to_hex(color)
            if (
                "supercategory" not in catg.keys()
                or catg["supercategory"] == "none"
                or catg["supercategory"] is None
                or len(catg["supercategory"]) == 0
            ):
                self.add_label(
                    name=catg["name"],
                    id=catg["id"],
                    super_category_id=None,
                    color=color,
                )
            else:
                super_category_id = self.label_name2label_id(catg["supercategory"])
                if super_category_id is None and catg["name"] not in tried_names:
                    catgs.append(catg)
                    tried_names.append(catg["name"])
                else:
                    self.add_label(
                        name=catg["name"],
                        id=catg["id"],
                        super_category_id=super_category_id,
                        color=color,
                    )
            db.session.commit()



import random
import time


def rgb_to_hex(rgb):
    rgb = tuple(int(c) for c in rgb)
    return ("#%02x%02x%02x" % rgb).upper()


def hex_to_rgb(value):
    value = value.lstrip("#")
    lv = len(value)
    return tuple(int(value[i : i + lv // 3], 16) for i in range(0, lv, lv // 3))


def name_to_hex(name):
    color_map = {
        "red": "#FF0000",
        "white": "#FFFFFF",
        "cyan": "#00FFFF",
        "silver": "#C0C0C0",
        "blue": "#0000FF",
        "gray": "#808080",
        "darkblue": "#0000A0",
        "black": "#000000",
        "lightblue": "#ADD8E6",
        "orange": "#FFA500",
        "purple": "#800080",
        "brown": "#A52A2A",
        "yellow": "#FFFF00",
        "maroon": "#800000",
        "lime": "#00FF00",
        "green": "#008000",
        "magenta": "#FF00FF",
        "olive": "#808000",
    }
    if name[0] == "#" and len(name) == 7:
        return name

    name = name.lower()
    if name not in color_map.keys():
        raise RuntimeError(f"Don't have hex value for color {name}")

    return color_map[name]


# TODO: generate color with high contrast
# TODO: generate color within certain range
# TODO: even devide color space
def rand_hex_color(current_colors=[]):
    random.seed(time.time())
    r = lambda: random.randint(0, 255)
    while True:
        c = "#%02X%02X%02X" % (r(), r(), r())
        c = c.upper()
        if c not in current_colors:
            break
    return c


# Red 	#FF0000
# White 	#FFFFFF
# Cyan 	#00FFFF
# Silver 	#C0C0C0
# Blue 	#0000FF
# Gray 	#808080
# DarkBlue 	#0000A0
# Black 	#000000
# LightBlue 	#ADD8E6
# Orange 	#FFA500
# Purple 	#800080
# Brown 	#A52A2A
# Yellow 	#FFFF00
# Maroon 	#800000
# Lime 	#00FF00
# Green 	#008000
# Magenta 	#FF00FF
# Olive 	#808000



import os
import os.path as osp
import shutil


# TODO: switch to pathlib

image_extensions = [".bmp", ".jpg", ".jpeg", ".png", ".gif", ".webp"]


def create_dir(path):
    if path is None:
        raise RuntimeError("Path to create is None")
    if not osp.isabs(path):
        raise RuntimeError(f"Only supports absolute path, got {path}")
    os.makedirs(path, exist_ok=True)


def remove_dir(path):
    if not osp.exists(path):
        return
    shutil.rmtree(path)


def listdir(folder, filters={"exclude_prefix": ["."]}):
    """
    return relative path of all files satisfying filters under the folder and its subfolders

    Args:
        folder (str): the folder to list
        filters (dict, optional): Four lists, include/exclude_prefix/postfix. Include first, satisfying either include, then exclude fail either one gets excluded.

    Returns:
        list: File paths relative to folder, sorted
    """

    files = []
    for root, fdrs, fs in os.walk(folder):
        if osp.basename(root).startswith("."):  # skip all hidden folders
            continue
        for f in fs:
            # files.append(osp.normpath(osp.join(root, f)))
            files.append(osp.join(root, f))
    files = [osp.relpath(f, folder) for f in files]
    # TODO: support regx
    include_prefix = filters.get("include_prefix", [])
    include_postfix = filters.get("include_postfix", [])

    def include(path):
        f = osp.basename(path)
        for pref in include_prefix:
            if f[: len(pref)].lower() == pref:
                return True
        for postf in include_postfix:
            if f[-len(postf) :].lower() == postf:
                return True
        return False

    if len(include_prefix) != 0 or len(include_postfix) != 0:
        files = list(filter(include, files))

    exclude_prefix = filters.get("exclude_prefix", [])
    exclude_postfix = filters.get("exclude_postfix", [])

    def exclude(path):
        f = osp.basename(path)
        for pref in exclude_prefix:
            if f[: len(pref)] == pref:
                return False
        for postf in exclude_postfix:
            if f[-len(postf) :] == postf:
                return False
        return True

    files = list(filter(exclude, files))
    files.sort()
    # files = [osp.normpath(p) for p in files]
    return files


def copy(src, dst, make_dir=False):
    # src = osp.normpath(src)
    # dst = osp.normpath(dst)
    if dst == osp.dirname(src):
        return
    if make_dir:
        os.makedirs(osp.dirname(dst), exist_ok=True)
    shutil.copy(src, dst)


def copycontent(src, dst):
    """
    Recursively copy everything in src to dst. Create dst if not exist.

    Args:
        src (str): source folder
        dst (str): destination folder
    """

    assert osp.abspath(src), f"src dir {src} isn't abspath"
    assert osp.abspath(dst), f"dst dir {dst} isn't abspath"
    assert src != dst, f"The source and destination folder are both {src}"

    for root, fdrs, fs in os.walk(src):
        if osp.basename(root).startswith("."):  # skip all hidden folders
            continue
        if not osp.exists(osp.join(dst, osp.relpath(root, src))):
            os.makedirs(osp.join(dst, osp.relpath(root, src)))

        for f in fs:
            fsrc = osp.join(root, f)
            fdst = osp.join(osp.join(dst, osp.relpath(root, src), f))
            if osp.exists(fdst):
                continue
            copy(fsrc, fdst)



import os.path as osp

from flask_cors import CORS
import flask

import paddlelabel
from paddlelabel.util import Resolver
from paddlelabel.config import sqlite_url, db, connexion_app, db_path
import paddlelabel.api
import paddlelabel.task


@connexion_app.app.route("/")
def index():
    return flask.redirect("/static/index.html")


if not osp.exists(db_path):
    print("Creating db")
    db.create_all()

    # TODO: move to base
    from paddlelabel.config import basedir
    from paddlelabel.api.controller.setting import init_site_settings

    init_site_settings(osp.join(basedir, "default_setting.json"))


connexion_app.add_api(
    "openapi.yml",
    resolver=Resolver("paddlelabel.api", collection_endpoint_name="get_all"),
    # request with undefined param returns error, dont enforce body
    strict_validation=True,
    pythonic_params=True,
)

CORS(connexion_app.app)



import json

from marshmallow import pre_load, post_dump, fields

from paddlelabel.api.model import Project
from .base import BaseSchema


class ProjectSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Project
        # include_relationships = True

    label_dir = fields.String(allow_none=True)
    task_category = fields.Nested("TaskCategorySchema")
    labels = fields.List(fields.Nested("LabelSchema"))

    @pre_load
    def pre_load_action(self, data, **kwargs):
        if "label_dir" in data.keys() and data["label_dir"] == "":
            data["label_dir"] = None

        if "other_settings" in data.keys():
            data["other_settings"] = json.dumps(data["other_settings"])
        return data

    @post_dump
    def pre_dump_action(self, project, **kwargs):
        if "other_settings" in project.keys() and project["other_settings"] is not None:
            project["other_settings"] = json.loads(project["other_settings"])
        # print(project["other_settings"])
        return project



import os
import os.path as osp

import connexion
from flask_sqlalchemy import SQLAlchemy  # TODO: remove
from flask_marshmallow import Marshmallow  # TODO: remove
from flask_cors import CORS

from .util import rand_string

basedir = os.path.abspath(os.path.dirname(__file__))

db_path = f"{osp.join(os.path.expanduser('~'), '.paddlelabel', 'paddlelabel.db')}"

if not osp.exists(osp.dirname(db_path)):
    os.makedirs(osp.dirname(db_path))
sqlite_url = f"sqlite:///{db_path}"
print("database url: ", sqlite_url)

connexion_app = connexion.App(__name__)
app = connexion_app.app
app.config["SQLALCHEMY_DATABASE_URI"] = sqlite_url
app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
app.config["SQLALCHEMY_ECHO"] = False
app.config["SECRET_KEY"] = rand_string(30)
app.static_url_path = "/static"
app.static_folder = osp.join(basedir, "static")
CORS(connexion_app.app)

db = SQLAlchemy(app)
se = db.session
ma = Marshmallow(app)


# reject requests with the same request_id within request_id_timeout seconds
request_id_timeout = 2

data_base_dir = osp.join(os.path.expanduser("~"), ".paddlelabel")



import os
import argparse

from .serve import connexion_app
from paddlelabel.api.controller.sample import prep_samples

parser = argparse.ArgumentParser(description="PP Label")
parser.add_argument(
    "--lan",
    "-l",
    default=False,
    action="store_true",
    help="Whether to expose the service to lan",
)
parser.add_argument(
    "--port",
    "-p",
    default=17995,
    type=int,
    help="The port to use",
)
args = parser.parse_args()


def run():
    prep_samples()
    host = "0.0.0.0" if args.lan else "127.0.0.1"
    connexion_app.run(host=host, port=args.port, debug=True)


run()



from datetime import datetime

from paddlelabel.config import db
from ..util import nncol
from .base import BaseModel


class Data(BaseModel):
    __tablename__ = "data"
    __table_args__ = {"comment": "Contains all the data files"}
    data_id = nncol(db.Integer(), primary_key=True)
    task_id = db.Column(db.Integer(), db.ForeignKey("task.task_id", ondelete="CASCADE"))
    # task = db.relationship("Task", lazy="selectin")
    annotations = db.relationship("Annotation", lazy="selectin", cascade="all, delete-orphan")
    path = nncol(db.String())
    size = db.Column(db.String())  # , seperated string. natural image: NHWC (C is optional)

    _immutables = BaseModel._immutables + ["data_id", "task_id"]



from marshmallow import fields, post_dump

from paddlelabel.api.model import Data
from .base import BaseSchema
from paddlelabel.api.schema.util import str2sault


class DataSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Data

    @post_dump
    def post_dump_action(self, data, **kwargs):
        data["sault"] = str2sault(data["path"] + data["created"])
        return data

    # task = fields.Nested("TaskSchema", exclude=("datas", "annotations"))



from .project import ProjectSchema
from .label import LabelSchema
from .task import TaskSchema
from .data import DataSchema
from .annotation import AnnotationSchema
from .user import UserSchema
from .setting import TaskCategorySchema
from .tag import TagSchema
from .tag_task import TagTaskSchema



import os.path as osp

import paddlelabel
from paddlelabel.config import sqlite_url, db

if not osp.exists(sqlite_url):
    print("Creating db")
    db.create_all()

    # TODO: move to base
    from paddlelabel.config import basedir
    from paddlelabel.api.controller.setting import init_site_settings

    init_site_settings(osp.join(basedir, "default_setting.json"))

# paddlelabel.task.classification.test()

paddlelabel.task.classification.single_clas()
# paddlelabel.task.classification.multi_clas()
#
# paddlelabel.task.detection.voc()
# paddlelabel.task.detection.coco()



from datetime import datetime

from paddlelabel.config import db
from paddlelabel.api.util import nncol
from .base import BaseModel


class Annotation(BaseModel):
    __tablename__ = "annotation"
    __table_args__ = {"comment": "Contains all the annotations"}
    annotation_id = nncol(db.Integer(), primary_key=True)
    frontend_id = db.Column(db.Integer())  # unique id within data, start from 1
    result = db.Column(db.String())
    type = db.Column(db.String())
    label_id = nncol(db.Integer(), db.ForeignKey("label.label_id", ondelete="CASCADE"))
    data_id = db.Column(db.Integer(), db.ForeignKey("data.data_id", ondelete="CASCADE"))

    # TODO: remove
    task_id = nncol(db.Integer(), db.ForeignKey("task.task_id", ondelete="CASCADE"))
    # task = db.relationship("Task")
    project_id = nncol(db.Integer(), db.ForeignKey("project.project_id"))

    _immutables = BaseModel._immutables + ["annotation_id", "label_id", "task_id", "slice_id"]



import math
import random
import json
import requests
import os
import os.path as osp
import base64
from copy import deepcopy
import traceback

import numpy as np
import connexion

from paddlelabel.config import db
from paddlelabel.api.model import Project, Task, TaskCategory, Annotation, Label, project
from paddlelabel.api.schema import ProjectSchema
from paddlelabel.api.controller.base import crud
from paddlelabel.api.controller import label
from paddlelabel.api.util import abort
from paddlelabel.task.util import rand_hex_color
from paddlelabel.util import camel2snake
from paddlelabel.task.util.file import (
    image_extensions,
    listdir,
    create_dir,
    remove_dir,
    copy,
)
import paddlelabel


def pre_add(new_project, se):
    if not osp.isabs(new_project.data_dir):
        abort("Dataset Path is not absolute path", 409)

    new_project.label_format = camel2snake(new_project.label_format)
    new_labels = new_project.labels
    rets, unique = label.unique_within_project(new_project.project_id, new_labels)
    if not np.all(unique):
        # TODO: return the not unique field
        abort("Project labels are not unique", 409)
    return new_project


def _import_dataset(project, data_dir=None):
    task_category = TaskCategory._get(task_category_id=project.task_category_id)

    # 1. create handler
    if task_category is None:
        handler = paddlelabel.task.BaseTask(project)
    else:
        handler = eval(task_category.handler)(project, data_dir=data_dir)

    # 2. choose importer. if specified, use importer for new_project.label_format, else use default_importer
    if project.label_format is not None:
        if project.label_format not in handler.importers.keys():
            abort(
                f"Importer {project.label_format} for project category {task_category.name} not found",
                404,
                "No such importer",
            )
        importer = handler.importers[project.label_format]
    else:
        importer = handler.default_importer

    # 3. run import
    importer(data_dir)


def post_add(new_project, se):
    """run task import after project creation"""
    try:
        _import_dataset(new_project)
    except Exception as e:
        project = Project.query.filter(Project.project_id == new_project.project_id).one()
        db.session.delete(project)
        db.session.commit()

        print("Create project failed")
        print(traceback.format_exc())

        if "detail" in dir(e):
            abort(e.detail, 500, e.title)
        else:
            abort(str(e), 500, str(e))

    return new_project


def export_dataset(project_id):
    _, project = Project._exists(project_id)
    task_category = TaskCategory._get(task_category_id=project.task_category_id)
    handler = eval(task_category.handler)(project)
    if project.label_format is not None:
        exporter = handler.exporters[project.label_format]
    else:
        exporter = handler.default_exporter
    req = connexion.request.json

    try:
        exporter(req["export_dir"])
    except Exception as e:
        abort(str(e), 500, str(e))


def import_dataset(project_id):
    """import additional data

    Args:
        project_id (int): the project to import to
    """
    # 1. get project
    req = connexion.request.json
    _, project = Project._exists(project_id)

    # 2. get current project data names
    tasks = Task._get(project_id=project.project_id, many=True)
    curr_data_names = set()
    for task in tasks:
        for data in task.datas:
            curr_data_names.add(osp.basename(data.path))

    # 3. move all new images and all other files to import temp

    import_temp = osp.join(osp.expanduser("~"), ".paddlelabel", "import_temp")
    remove_dir(import_temp)
    create_dir(import_temp)

    import_dir = req["import_dir"]
    new_data_paths = listdir(
        import_dir,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    )
    all_paths = listdir(
        import_dir,
        filters={"exclude_prefix": ["."]},
    )
    all_copy_paths = [p for p in all_paths if p not in new_data_paths]
    new_data_paths = [p for p in new_data_paths if osp.basename(p) not in curr_data_names]
    all_copy_paths += new_data_paths
    # print(all_copy_paths)
    for p in all_copy_paths:
        copy(osp.join(import_dir, p), osp.join(import_temp, p), make_dir=True)

    _import_dataset(project, import_temp)

    for p in new_data_paths:
        copy(osp.join(import_temp, p), osp.join(project.data_dir, p), make_dir=True)

    # remove_dir(import_temp)


def pre_put(project, body, se):
    if "other_settings" in body.keys():
        body["other_settings"] = json.dumps(body["other_settings"])
    return project, body


def split_dataset(project_id, epsilon=1e-3):
    Project._exists(project_id)
    split = connexion.request.json
    if list(split.keys()) != ["train", "val", "test"]:
        abort(
            f"Got {split}",
            500,
            "Request should provide train, validataion and test percentage",
        )  # TODO: change response code
    if abs(1 - sum(split.values())) > epsilon:
        abort(
            f"The train({split['train']}), val({split['val']}), test({split['test']}) split don't sum to 1.",
            500,
            "The three percentages don't sum to 1",
        )  # TODO: change response code
    split_num = [0] * 4
    split_num[1] = split["train"]
    split_num[2] = split["val"]
    split_num[3] = split["test"]
    split = split_num
    for idx in range(1, 4):
        split[idx] += split[idx - 1]

    tasks = Task._get(project_id=project_id, many=True)
    split = [math.ceil(s * len(tasks)) for s in split]
    # print("split numbers: ", len(tasks), split)
    random.shuffle(tasks)
    for set in range(3):
        for idx in range(split[set], split[set + 1]):
            tasks[idx].set = set
    db.session.commit()
    tasks = Task._get(project_id=project_id, many=True)
    return {
        "train": split[1],
        "val": split[2] - split[1],
        "test": split[3] - split[2],
    }, 200


def create_label(project, label_name):
    color = rand_hex_color([l.color for l in project.labels])
    ids = [l.id for l in project.labels]
    ids.append(0)
    label = Label(
        id=max(ids) + 1,
        project_id=project.project_id,
        name=label_name,
        color=color,
    )
    project.labels.append(label)
    db.session.commit()
    return label


def predict(project_id):
    _, project = Project._exists(project_id)

    params = connexion.request.json
    if "create_label" not in params.keys():
        params["create_label"] = False
    if "same_server" not in params.keys():
        params["same_server"] = False

    url = params["ml_backend_url"]
    if url[-1] != "/":
        url += "/"
    url += params["model"] + "/predict"
    # print("request url", url)

    headers = {"content-type": "application/json"}

    labels = Label._get(project_id=project_id, many=True)
    labels = {l.name: l.label_id for l in labels}

    for task in Task._get(project_id=project_id, many=True):
        for data in task.datas:
            if len(data.annotations) != 0:
                continue
            if params["same_server"]:
                body = {"img": osp.join(project.data_dir, data.path), "format": "path"}
            else:
                img_b64 = base64.b64encode(open(osp.join(project.data_dir, data.path), "rb").read()).decode("utf-8")
                body = {"img": img_b64, "format": "b64"}
            res = requests.post(url, headers=headers, json=body)
            res = json.loads(res.text)
            if res["result"] not in labels.keys():
                if params["create_label"]:
                    new_label = create_label(project, res["result"])
                    labels[new_label.name] = new_label.id
                else:
                    continue
            ann = Annotation(
                label_id=labels[res["result"]],
                project_id=project.project_id,
                # task_id=task.task_id,
                data_id=data.data_id,
                result="",
            )
            task.annotations.append(ann)
            # print(osp.join(project.data_dir, data.path), res["result"])
    db.session.commit()
    return "finished"


def post_delete(project, se):
    warning_path = osp.join(project.data_dir, "paddlelabel.warning")
    if osp.exists(warning_path):
        os.remove(warning_path)


get_all, get, post, put, delete = crud(
    Project,
    ProjectSchema,
    triggers=[pre_add, post_add, pre_put, post_delete],
)



import json
import os
import os.path as osp
import random

import cv2
import flask
import tempfile

from paddlelabel.config import db
from .base import crud
from ..model import Data, Project, Task
from ..schema import DataSchema
from paddlelabel.api.util import abort
from paddlelabel.task.segmentation import draw_mask

get_all, get, post, put, delete = crud(Data, DataSchema)


# TODO: dont use flask
def get_image(data_id):
    # if random.random()<0.9:
    #     abort("Mimic package loss", 404)
    _, data = Data._exists(data_id)
    path = data.path
    project_id = data.task.project_id
    data_dir = Project._get(project_id=project_id).data_dir

    folder = osp.join(data_dir, osp.dirname(path))
    file_name = osp.basename(path)
    # return flask.send_from_directory(data_dir, path)

    return flask.send_from_directory(folder, file_name)

    # data_path = osp.join(data_dir, path)
    # image = cv2.imread(data_path)
    # image_png = cv2.imencode(".png", image)
    # b64_string = base64.b64encode(image_png[1]).decode("utf-8")
    # return json.dumps({"image": b64_string}), 200


def get_mask(data_id):
    _, data = Data._exists(data_id)
    mask = draw_mask(data)
    if mask is None:
        abort("This data probably doesn't have segmentation mask", 500)

    tempf = tempfile.NamedTemporaryFile(suffix=".png")
    cv2.imwrite(tempf.name, mask)

    return flask.send_from_directory(osp.dirname(tempf.name), osp.basename(tempf.name))


def get_by_task(task_id):
    Task._exists(task_id)
    datas = Data._get(task_id=task_id, many=True)
    return DataSchema(many=True).dump(datas), 200



from marshmallow import fields


from paddlelabel.api.model import Annotation
from .base import BaseSchema


class AnnotationSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Annotation

    # not required when calling api
    project_id = fields.Integer()
    task_id = fields.Integer()
    # task = fields.Nested("TaskSchema", exclude=("annotations", "datas"))
    label = fields.Nested("LabelSchema")



from datetime import datetime

from sqlalchemy.orm import backref

from paddlelabel.config import db
from ..util import nncol
from .annotation import Annotation
from .project import Project
from .base import BaseModel


class Task(BaseModel):
    __tablename__ = "task"
    __table_args__ = {"comment": "Contains all the tasks"}
    task_id = nncol(db.Integer(), primary_key=True)
    project_id = nncol(
        db.Integer(),
        db.ForeignKey("project.project_id", ondelete="CASCADE"),
    )
    datas = db.relationship("Data", lazy="selectin", backref="task", cascade="all, delete-orphan")
    annotations = db.relationship("Annotation", lazy="selectin", backref="task", cascade="all, delete-orphan")
    # TODO: split with tag!
    set = nncol(db.Integer())  # 0 train, 1 val, 2 test



import time

from flask import abort, session, request

import paddlelabel
from paddlelabel.config import app, request_id_timeout

from . import controller
from .model import Project, Label, Task, Data, Annotation, TaskCategory
from .util import decode_token
from . import rpc

# TODO: is this specific to flask or is connexion function
@app.before_request
def check_request_id():
    request_id = request.headers.get("request_id", None)
    if request_id is None or len(request_id) == 0:
        return
    curr_time = time.time()
    if "request_ids" not in session:
        session["request_ids"] = [(curr_time, request_id)]
        return
    session["request_ids"] = list(
        filter(
            lambda item: curr_time - item[0] < request_id_timeout,
            session["request_ids"],
        )
    )
    exist = False
    for ts, id in session["request_ids"]:
        if id == request_id:
            exist = True
    session["request_ids"].append((curr_time, request_id))
    if exist:
        abort(409, f"Duplicate request from {curr_time - ts}s ago")



import json

from ..model import TaskCategory

from paddlelabel.config import se


# TODO: change task_categories to project_category
def init_site_settings(json_path):
    settings = json.loads(open(json_path, "r").read())
    task_categories = settings["site"]["task_categories"]
    for idx, (cat, handler) in task_categories.items():
        idx = int(idx)
        curr_cat = TaskCategory.query.filter(TaskCategory.task_category_id == idx).one_or_none()
        if curr_cat is None:
            curr_cat = TaskCategory(idx, cat, handler)
            se.add(curr_cat)
            se.commit()



from marshmallow import pre_load, fields

from paddlelabel.api.model import User
from .base import BaseSchema


class UserSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = User

    uuid = fields.String()



from datetime import datetime

from paddlelabel.config import db
from ..util import nncol
from .base import BaseModel


class Label(BaseModel):
    __tablename__ = "label"
    __table_args__ = {"comment": "Contains all the label information"}
    label_id = nncol(db.Integer(), primary_key=True)
    id = nncol(db.Integer())
    project_id = db.Column(db.Integer(), db.ForeignKey("project.project_id"))  # TODO: why missing project_id when nncol
    name = nncol(db.String())
    color = db.Column(db.String())
    comment = db.Column(db.String())
    annotations = db.relationship("Annotation", lazy="noload", backref="label")  # cascade="all, delete-orphan"
    super_category_id = db.Column(db.Integer())  # TODO: foreign key
    _immutables = BaseModel._immutables + ["label_id", "project_id"]



import os.path as osp

from functools import total_ordering

from paddlelabel.config import db
from paddlelabel.api.util import nncol
from .base import BaseModel


class TaskCategory(BaseModel):
    __tablename__ = "taskCategory"
    task_category_id = nncol(
        db.Integer(),
        primary_key=True,
        autoincrement=False,
    )
    name = nncol(db.String(), unique=True)
    handler = db.Column(db.String())

    def __init__(self, task_category_id, name, handler):
        super().__init__()
        self.task_category_id = task_category_id
        self.name = name
        if len(handler) == 0:
            self.handler = None
        else:
            self.handler = handler



from marshmallow import fields

from paddlelabel.api.model import Label
from .base import BaseSchema


class LabelSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Label

    id = fields.Integer()



import connexion

from .base import crud
from ..model import Annotation, Task, Project, Data
from ..schema import AnnotationSchema
from ..util import abort

from paddlelabel.config import db

# TODO: don't create label if exists.


def pre_add(annotation, se):
    _, data = Data._exists(annotation.data_id)
    _, task = Task._exists(data.task_id)
    annotation.task_id = data.task_id
    annotation.project_id = task.project_id

    return annotation


get_all, get, post, put, delete = crud(Annotation, AnnotationSchema, [pre_add])


def get_by_project(project_id):
    Project._exists(project_id)
    anns = Annotation._get(project_id=project_id, many=True)
    return AnnotationSchema(many=True).dump(anns), 200


def get_by_task(task_id):
    Task._exists(task_id)
    anns = Annotation._get(task_id=task_id, many=True)
    return AnnotationSchema(many=True).dump(anns), 200


def get_by_data(data_id):
    Data._exists(data_id)
    anns = Annotation._get(data_id=data_id, many=True)
    return AnnotationSchema(many=True).dump(anns), 200


def set_all_by_data(data_id):
    """Receive all annotataions under a data record from front end and update database

    Args:
        data_id (int): data_id
    """

    # 1. ensure data exists
    _, data = Data._exists(data_id)

    # 2. load and parse all new anns
    schema = AnnotationSchema()
    anns = connexion.request.json
    for idx in range(len(anns)):
        del anns[idx]["label"]
        anns[idx] = schema.load(anns[idx])
    curr_anns = Annotation._get(data_id=data.data_id, many=True)  # query before the first edit
    curr_ann_ids = set([ann.annotation_id for ann in curr_anns])

    # 3. add new ann and update existing ann
    task = Task._get(task_id=data.task_id)
    keep_ann_ids = []  # anns to keep, newly created anns don't need to be included in this
    for ann in anns:
        # if ann doesn't have id or id not in curr ids, it's new
        if ann.annotation_id is None or ann.annotation_id not in curr_ann_ids:
            ann.annotation_id = None
            ann.task_id = task.task_id
            ann.project_id = task.project_id
            db.session.add(ann)
        else:
            keep_ann_ids.append(ann.annotation_id)
            ann_dict = schema.dump(ann)
            del ann_dict["created"]
            del ann_dict["modified"]
            del ann_dict["annotation_id"]
            del ann_dict["label"]
            Annotation.query.filter(Annotation.annotation_id == ann.annotation_id).update(ann_dict)

    # 4. remove anns that are in db, but not in anns
    # print("curr_anns", curr_anns)
    for curr_ann in curr_anns:
        if curr_ann.annotation_id not in keep_ann_ids:
            db.session.delete(curr_ann)

    db.session.commit()

    curr_anns = Annotation._get(data_id=data.data_id, many=True)
    schema = AnnotationSchema(many=True)
    return schema.dump(curr_anns), 200


def delete_by_data(data_id):
    anns = Annotation._get(data_id=data_id, many=True)
    for ann in anns:
        db.session.delete(ann)
    db.session.commit()


def set_all_by_task_and_frontend(task_id):
    _, task = Task._exists(task_id)
    # 删除task关联的所有annotation
    current_anns = Annotation._get(task_id=task_id, many=True)
    for cann in current_anns:
        db.session.delete(cann)
    # 插入新annotation
    anns = connexion.request.json
    schema = AnnotationSchema()
    for ann in anns:
        ann = schema.load(ann)
        ann.task_id = task.task_id
        ann.project_id = task.project_id
        task.annotations.append(ann)
    db.session.commit()



import functools
import time

import connexion
import sqlalchemy as sa

from paddlelabel.util import camel2snake

# sa default to nullable=True
nncol = functools.partial(sa.Column, nullable=False)

# TODO: settle on how to use detail and title
def abort(detail: str, status: int, title: str = ""):
    detail = detail.replace("\n", " ")
    title = title.replace("\n", " ")
    raise connexion.exceptions.ProblemException(
        detail=detail,
        title=title if len(title) != 0 else detail,
        status=status,
        headers={"detail": detail},
    )


# TODO: move to config
JWT_ISSUER = "paddlelabel"
JWT_SECRET = "change_this"
JWT_LIFETIME_SECONDS = 43200  # 12h
JWT_ALGORITHM = "HS256"


def generate_token(uuid):
    timestamp = int(time.time())
    payload = {
        "iss": JWT_ISSUER,
        "iat": timestamp,
        "exp": timestamp + JWT_LIFETIME_SECONDS,
        "sub": uuid,
    }
    return jwt.encode(payload, JWT_SECRET, algorithm=JWT_ALGORITHM)


def decode_token(token):
    try:
        return jwt.decode(token, JWT_SECRET, algorithms=[JWT_ALGORITHM])
    except JWTError as e:
        raise Unauthorized from e


def parse_order_by(modal, order_by):
    order_by = order_by.split(" ")
    if len(order_by) == 2:
        key, sort_dir = order_by
        # print(key, sort_dir)
        if "asc" in sort_dir:
            sort_dir = "asc"
        else:
            sort_dir = "desc"
    else:
        key = order_by[0]
        sort_dir = "acs"
    key = camel2snake(key)

    try:
        order = getattr(getattr(modal, key), sort_dir)()
    except:
        order = modal.created.asc()
    return order



import re

from marshmallow import fields
import uuid
from werkzeug.security import generate_password_hash, check_password_hash

from ..model import User
from ..schema import UserSchema
from .base import crud
from ..util import abort, generate_token


def validata_email(email):
    regex = re.compile(r"([A-Za-z0-9]+[.-_])*[A-Za-z0-9]+@[A-Za-z0-9-]+(\.[A-Z|a-z]{2,})+")
    if re.fullmatch(regex, email):
        return True
    return False


def pre_add(user, se):
    if not validata_email(user.email):
        abort("Email is not valid", 400)
    user.password = generate_password_hash(user.password, method="sha256")
    user.uuid = str(uuid.uuid4())
    return user


def pre_put(curr_user, new_user, se):
    if "email" in new_user.keys():
        if not validata_email(new_user["email"]):
            abort("Email is not valid", 400)
    if "password" in new_user.keys():
        new_user["password"] = generate_password_hash(new_user["password"], method="sha256")


get_all, get, post, put, delete = crud(User, UserSchema, [pre_add, pre_put])


def login():
    r = connexion.request.json
    username = r.get("username", None)
    password = r.get("password", None)
    if username is None:
        abort("Need username to login", 401)
    if password is None:
        abort("Need password to login", 401)
    user = User._get(username=username)
    if user is None:
        abort(f"Username or password is wrong", 401)
    if check_password_hash(user.password, password):
        return generate_token(user.uuid)
    else:
        abort(f"Username or password is wrong", 401)



import math

import numpy as np
from PIL import Image, ImageDraw
import connexion
from .util.polygon import mask2polygon

import matplotlib.pyplot as plt


def polygon2mask(poly):
    """Convert polygon to mask.

    Origion at top left, width height.

    Args:
        poly (list): - list of points: [(w, h), (w, h), ..., (w, h)]
                     - list of flattened points: [w, h, w, h, ..., w, h]
    """

    poly = np.array(poly)

    if poly.ndim == 1:
        poly = poly.reshape((-1, 2))

    wmin, hmin = poly.min(axis=0)
    wmax, hmax = poly.max(axis=0)
    width = wmax - wmin + 1
    height = hmax - hmin + 1

    poly[:, 0] -= wmin
    poly[:, 1] -= hmin

    poly = [(p[0], p[1]) for p in poly.tolist()]

    img = Image.new("L", [width, height], 0)
    ImageDraw.Draw(img).polygon(poly, outline=1, fill=1)  # col, row
    mask = np.array(img) == 1
    # plt.imshow(mask)
    # plt.show()
    # print("=-=-=-=-=-=", width, height, mask.shape)

    # plt.imshow(mask)
    # plt.show()
    # print(mask)

    return (wmin, hmin), mask


# polygon2mask([(1, 1), (2, 1), (3, 1), (2, 8)])
# polygon2mask([(0,0), (100, 0), (100, 20), (0, 30)])
# polygon2mask([0, 0, 100, 0, 100, 20, 0, 30])


"""
	{
		"annotation_id": 2,
		"created": "2022-06-11T11:24:42.484401",
		"data_id": 3,
		"frontend_id": null,
		"label": {
			"color": "#800000",
			"comment": null,
			"created": "2022-06-11T11:24:42.286857",
			"id": 1,
			"label_id": 1,
			"modified": "2022-06-11T11:24:42.286861",
			"name": "视盘",
			"project_id": 1,
			"super_category_id": null
		},
		"label_id": 1,
		"modified": "2022-06-11T11:24:42.484404",
		"project_id": 1,
		"result": "0,1,30,10,31,10,32,10,33,10,34,10,35,10,36,10,37,10,38,10,39,10,30,11,31,11,32,11,33,11,34,11,35,11,36,11,37,11,38,11,39,11,30,12,31,12,32,12,33,12,34,12,35,12,36,12,37,12,38,12,39,12,30,13,31,13,32,13,33,13,34,13,35,13,36,13,37,13,38,13,39,13,30,14,31,14,32,14,33,14,34,14,35,14,36,14,37,14,38,14,39,14,30,15,31,15,32,15,33,15,34,15,35,15,36,15,37,15,38,15,39,15,30,16,31,16,32,16,33,16,34,16,35,16,36,16,37,16,38,16,39,16,30,17,31,17,32,17,33,17,34,17,35,17,36,17,37,17,38,17,39,17,30,18,31,18,32,18,33,18,34,18,35,18,36,18,37,18,38,18,39,18,30,19,31,19,32,19,33,19,34,19,35,19,36,19,37,19,38,19,39,19",
		"task_id": 3,
		"type": "brush"
	},
"""


def polygon2points(poly):
    (wmin, hmin), mask = polygon2mask(poly)
    points = []
    for idh in range(mask.shape[0]):
        for idw in range(mask.shape[1]):
            if mask[idh][idw] == 1:
                points.append((wmin + idw, hmin + idh))

    return points


# print(polygon2points([(0, 1), (3, 2), (3, 3), (1, 3)]))
# print(polygon2points([0, 1, 3, 2, 3, 3, 1, 3]))


def polygon2points_str():
    poly = connexion.request.json["polygon"]
    # print(poly, type(poly), type(poly[0]))
    poly = poly.split(",")
    poly = [int(p) for p in poly]

    (wmin, hmin), mask = polygon2mask(poly)
    points = []
    for idw in range(mask.shape[0]):
        for idh in range(mask.shape[1]):
            if mask[idw][idh] == 1:
                points.append(str(wmin + idw))
                points.append(str(hmin + idh))

    return ",".join(points)


# print(polygon2points([(0, 1), (3, 2), (3, 3), (1, 3)]))
# print(polygon2points_str([0, 1, 3, 2, 3, 3, 1, 3]))


def points2polygon_str(ignore_first_two=False):
    points = connexion.request.json["points"]
    points = points.split(",")
    if ignore_first_two:
        points = points[2:]
    points = [int(p) for p in points]
    points = np.array(points)
    points = points.reshape((-1, 2))
    wmax, hmax = points.max(axis=0)
    mask = np.zeros((wmax + 1, hmax + 1), dtype="uint8")
    for p in points:
        mask[p[0], p[1]] = 255
    polygons = mask2polygon(mask)
    polygons = [[str(v) for point in poly for v in point] for poly in polygons]
    polygons = [",".join(poly) for poly in polygons]

    return polygons



import os
import os.path as osp

import connexion

from paddlelabel.config import data_base_dir
from paddlelabel.api.util import abort


def get_folders():
    path = connexion.request.json.get("path", None)
    if path is None:
        path = ""
    target = osp.join(data_base_dir, path)
    target = osp.abspath(target)
    # print(target, type(target))

    if not target[-1] == os.sep:
        target += os.sep
    if not target.startswith(osp.abspath(data_base_dir) + os.sep):
        abort("Can only request folders under data_base_dir", 500)
    if not osp.exists(target):
        return [], 200
    return next(os.walk(target))[1], 200



from paddlelabel.config import db
from paddlelabel.api.util import nncol
from .base import BaseModel


class Tag(BaseModel):
    __tablename__ = "tag"
    __table_args__ = {"comment": "Contains all the tags"}
    tag_id = nncol(db.Integer(), primary_key=True)
    project_id = nncol(db.Integer(), db.ForeignKey("project.project_id"))
    name = nncol(db.String())
    color = db.Column(db.String())
    comment = db.Column(db.String())

    _immutables = BaseModel._immutables + ["tag_id", "project_id"]



import os
import os.path as osp
import requests
import tarfile
import zipfile
from pathlib import Path

from tqdm import tqdm
import connexion
import flask

import paddlelabel
from paddlelabel.config import data_base_dir
from paddlelabel.api.schema import ProjectSchema
from paddlelabel.api.model import TaskCategory, Project
from paddlelabel.api.util import abort
from paddlelabel.config import basedir
from paddlelabel.task.util.file import copy, copycontent


def prep_samples(sample_dst: str = None):
    if sample_dst is None:
        sample_dst = osp.join(osp.expanduser("~"), ".paddlelabel", "sample")
    sample_source = osp.join(basedir, "sample")
    copycontent(sample_source, sample_dst)

    dsts = [
        "clas/multi/image/1.jpeg",
        "clas/multi/image/2.jpeg",
        "clas/multi/image/3.jpeg",
        "clas/multi/image/4.jpeg",
        "clas/single/1/1.jpeg",
        "clas/single/2/2.jpeg",
        "clas/single/3/3.jpeg",
        "clas/single/4/4.jpeg",
        "det/coco/JPEGImages/1.jpeg",
        "det/coco/JPEGImages/2.jpeg",
        "det/coco/JPEGImages/3.jpeg",
        "det/coco/JPEGImages/4.jpeg",
        "det/voc/JPEGImages/1.jpeg",
        "det/voc/JPEGImages/2.jpeg",
        "det/voc/JPEGImages/3.jpeg",
        "det/voc/JPEGImages/4.jpeg",
        "instance_seg/mask/JPEGImages/1.jpeg",
        "instance_seg/mask/JPEGImages/2.jpeg",
        "instance_seg/mask/JPEGImages/3.jpeg",
        "instance_seg/mask/JPEGImages/4.jpeg",
        "instance_seg/polygon/image/1.jpeg",
        "instance_seg/polygon/image/2.jpeg",
        "instance_seg/polygon/image/3.jpeg",
        "instance_seg/polygon/image/4.jpeg",
        "semantic_seg/mask/JPEGImages/1.jpeg",
        "semantic_seg/mask/JPEGImages/2.jpeg",
        "semantic_seg/mask/JPEGImages/3.jpeg",
        "semantic_seg/mask/JPEGImages/4.jpeg",
        "semantic_seg/polygon/image/1.jpeg",
        "semantic_seg/polygon/image/2.jpeg",
        "semantic_seg/polygon/image/3.jpeg",
        "semantic_seg/polygon/image/4.jpeg",
    ]
    img_path = osp.join(sample_source, "imgs")
    for dst in dsts:
        dst = osp.join(sample_dst, dst)
        src = osp.join(img_path, osp.basename(dst))
        copy(src, dst, make_dir=True)


def load_sample():
    prep_samples()

    task_category_id = connexion.request.json.get("task_category_id")

    sample_folder = {
        "classification": ["clas", "single"],
        "detection": ["det", "voc"],
        "semantic_segmentation": ["semantic_seg", "mask"],
        "instance_segmentation": ["instance_seg", "polygon"],
    }
    label_formats = {
        "classification": "single_class",
        "detection": "voc",
        "semantic_segmentation": "mask",
        "instance_segmentation": "polygon",
    }
    task_category = TaskCategory._get(task_category_id=task_category_id)
    data_dir = osp.join(osp.expanduser("~"), ".paddlelabel", "sample", *sample_folder[task_category.name])
    curr_project = Project._get(data_dir=data_dir)
    if curr_project is not None:
        # abort(f"Sample project for {task_category.name} is already created. Please visit home page to enter the project.", 500)
        return {"project_id": curr_project.project_id}, 200

    # print(task_category.name, task_category)

    project = {
        "name": f"Sample Project - {task_category.name}",
        "description": f"A {task_category.name} sample project created by PP-Label",
        "task_category_id": str(task_category_id),
        "data_dir": data_dir,
        "label_format": label_formats[task_category.name],
    }
    project = ProjectSchema().load(project)

    if task_category is None:
        handler = paddlelabel.task.BaseTask(project)
    else:
        handler = eval(task_category.handler)(project, data_dir=data_dir)

    handler.default_importer()

    # print(handler.project)

    return {"project_id": handler.project.project_id}, 200


def sample_folder_structure(path):
    # path = connexion.request.args['path']
    base_path = osp.join(osp.join(osp.expanduser("~"), ".paddlelabel", "sample"))
    path.replace("/", osp.sep)
    path = osp.join(base_path, path)

    def dfs(path):
        res = []
        names = os.listdir(path)
        for name in names:
            if name == "paddlelabel.warning":
                continue
            temp = {}
            full_path = osp.join(path, name)
            if osp.isdir(full_path):
                temp["title"] = name
                temp["key"] = osp.relpath(full_path, base_path)
                temp["children"] = dfs(full_path)
            else:
                temp["title"] = name
                temp["key"] = osp.relpath(full_path, base_path)
                temp["isLeaf"] = True
            res.append(temp)
        # res.sort()
        return res

    res = dfs(path)

    return res, 200


def serve_sample_file(path):
    base_path = osp.join(osp.join(osp.expanduser("~"), ".paddlelabel", "sample"))
    path.replace("/", osp.sep)
    path = osp.join(base_path, path)

    file_name = osp.basename(path)
    folder = osp.dirname(path)
    return flask.send_from_directory(folder, file_name)



import connexion

from paddlelabel.config import db
from .base import crud
from ..model import Tag, Project, Task, TagTask
from ..schema import TagSchema, TagTaskSchema
from ..util import abort


def pre_add(new_tag, se):
    curr_tag = Tag._get(project_id=new_tag.project_id, name=new_tag.name)
    if curr_tag is not None:
        abort(
            f"Duplicate tag with tag name {new_tag.name} under project {new_tag.project_id}",
            409,
        )
    return new_tag


get_all, get, post, put, delete = crud(Tag, TagSchema, [pre_add])


def get_by_project(project_id):
    Project._exists(project_id)
    tags = Tag._get(project_id=project_id, many=True)
    return TagSchema(many=True).dump(tags), 200


def get_by_task(task_id):
    Task._exists(task_id)
    tag_tasks = TagTask._get(task_id=task_id, many=True)
    tags = []
    for tag_task in tag_tasks:
        tag = Tag._get(tag_id=tag_task.tag_id)
        tags.append(tag)
    return TagSchema(many=True).dump(tags), 200


def add_to_task(task_id):
    # 1. check
    # 1.1 check task exists
    task = Task._get(task_id=task_id)
    if task is None:
        abort(f"No task with task id {task_id}", 404)

    body = connexion.request.json
    tag_id = body["tag_id"]
    # 1.2 check tag exist
    tag = Tag._get(tag_id=tag_id)
    if tag is None:
        abort(f"No tag with tag id {tag_id}", 404)
    # 1.3 no duplicate records
    curr_tag_task = TagTask._get(task_id=task_id, tag_id=tag_id)
    if curr_tag_task is not None:
        abort(f"Task {task_id} is already tagged {tag_id}", 409)
    # 1.4 check tag under task
    project_id = Task._get(task_id=task_id).project_id
    if tag.project_id != project_id:
        abort(f"Tag {tag_id} is not under project {project_id}", 404)

    tag_task = TagTask(
        tag_id=tag_id,
        task_id=task_id,
        project_id=task.project_id,
    )
    db.session.add(tag_task)
    db.session.commit()
    tag_tasks = TagTask._get(task_id=task_id, many=True)
    tags = []
    for tag_task in tag_tasks:
        tags.append(tag_task.tag)
    return TagSchema(many=True).dump(tags), 201



import random

import numpy as np
from sqlalchemy.sql.expression import func
import connexion

from paddlelabel.config import db
from paddlelabel.api.util import abort
from .base import crud
from ..model import Label, Project, Annotation
from ..schema import LabelSchema
from ..util import abort
from paddlelabel.task.util import rand_hex_color


# TODO: simplify with _get
def unique_within_project(project_id, new_labels=[], col_names=["id", "name"]):
    """check whether label id and name is unique within project

    Parameters
    ----------
    project_id : int
        The project labels belong to
    new_labels : [[], [], ..., []]
        The labels to be checked, each contains propreties of a new label
    col_names : ["", "", ..., ""]
        Column names to be checked

    Returns
    -------
    list
        [[True/False, ...,True/False], ..., [True/False, ...,True/False]]
        Whether each label is unique
    list
        [True/False, ...,True/False]
        When there's only one input
    """
    labels = Label.query.filter(Label.project_id == project_id).all()
    rets = []
    for column_idx, column_name in enumerate(col_names):
        results = [True] * len(new_labels)
        curr_values = set()
        for label in labels:
            curr_values.add(getattr(label, column_name))
        new_values = set()
        for idx, new_label in enumerate(new_labels):
            new_value = getattr(new_label, column_name)
            if new_value in curr_values:
                results[idx] = False
                continue
            if new_value in new_values:
                results[idx] = False
                continue
            new_values.add(new_value)
        rets.append(results)
    rets = np.array(rets)
    unique = np.all(rets, axis=0)

    if len(rets) == 1:
        return rets[0], unique[0]
    return rets, unique


def pre_add(new_label, se):
    # 1. label must have project_id and project must exist,
    if new_label.project_id is None:
        abort("Must specify project_id", 400)
    Project._exists(new_label.project_id)

    # 2. generate id, color
    if new_label.id is None:
        maxid = se.query(func.max(Label.id)).filter_by(project_id=new_label.project_id).one()[0]
        if maxid is None:
            maxid = 0
        new_label.id = maxid + 1
    if new_label.color is None:
        colors = Label.query.with_entities(Label.color).filter(Label.project_id == new_label.project_id).all()
        colors = [c[0] for c in colors]
        new_label.color = rand_hex_color(colors)

    # 3. cols must be unique within project
    cols = ["id", "name", "color"]
    rets, unique = unique_within_project(new_label.project_id, [new_label], cols)
    if not unique[0]:
        not_unique_cols = ", ".join([c for c, u in zip(cols, rets) if not u])
        abort(f"Label {not_unique_cols} is not unique", 409)

    return new_label


def in_use(label_id):
    annotations = Annotation.query.filter(Annotation.label_id == label_id).all()
    if len(annotations) == 0:
        return False
    return True


def pre_delete(label, se):
    if in_use(label.label_id):
        abort(f"Can't delete label {label.name} with annotation record", 409)
    sub_catgs = Label._get(super_category_id=label.label_id, many=True)
    print("sub_catgs", sub_catgs)
    if len(sub_catgs) != 0:
        abort(f"Can't delete label {label.name} which is super category to other labels", 409)
    return label


def get_by_project(project_id):
    Project._exists(project_id)
    labels = Label.query.filter(Label.project_id == project_id).all()
    return LabelSchema(many=True).dump(labels), 200


def delete_by_project(project_id, project_exists=False):
    if not project_exists:
        Project._exists(project_id)
    labels = Label._get(project_id=project_id, many=True)
    for lab in labels:
        db.session.delete(lab)
    db.session.commit()


def set_by_project(project_id):
    # _, project = Project._exists(project_id)
    # delete_by_project(project_id, project_exists=True)
    # schema = LabelSchema(many=True)
    # labels = schema.load(connexion.request.json)
    # for lab in labels:
    #     project.labels.append(lab)
    # db.session.commit()
    # return schema.dump(project.labels), 200
    abort("Not implemented", 500, "Not implemented")


get_all, get, post, put, delete = crud(Label, LabelSchema, triggers=[pre_add, pre_delete])


# TODO: abstract to any column

"""
label_id = nncol(db.Integer(), primary_key=True)
id = nncol(db.Integer())
project_id = db.Column(
    db.Integer(), db.ForeignKey("project.project_id", ondelete="CASCADE")
)
name = nncol(db.String())
color = db.Column(db.String())
comment = db.Column(db.String())
created = nncol(db.DateTime, default=datetime.utcnow)
modified = nncol(db.DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

- [x]. id unique within project
- [x]. name unique within project
- [x]. label with annotation in use cant be deleted

"""



from .file import create_dir, listdir, copy, copycontent, image_extensions
from .manager import ComponentManager
from .color import rgb_to_hex, hex_to_rgb, rand_hex_color, name_to_hex  # TODO: remove



from datetime import datetime

from paddlelabel.config import db
from paddlelabel.api.util import nncol
from ..util import abort

# TODO: nn string col cant be ""
class BaseModel(db.Model):
    __abstract__ = True
    __tablename__ = ""
    __table_args__ = {"comment": ""}
    created = nncol(db.DateTime, default=datetime.utcnow)
    modified = nncol(db.DateTime, default=datetime.utcnow, onupdate=datetime.utcnow)

    _immutables = ["created", "modified", "immutables"]
    _nested = ["project"]

    @classmethod
    @property
    def _cols(cls):
        return [c.key for c in cls.__table__.columns]

    def __repr__(self):
        s = f"Object {self.__tablename__}:  "
        for att in dir(self):
            if (
                att[0] != "_"
                and att[-1] != "s"
                and att not in ["query", "registry", "metadata", "query_class"]
                and att not in self._nested
            ):
                s += f"{att}: {getattr(self, att)}  "
        s += "\n"
        return s

    @classmethod
    def _exists(cls, item_id, throw=True):
        item = cls.query.filter(getattr(cls, cls.__tablename__ + "_id") == item_id).one_or_none()
        if item is None:
            if throw:
                abort(f"No {cls.__tablename__} with id : {item_id}", 404)
            else:
                return False, None
        return True, item

    @classmethod
    def _get(cls, **kwargs):
        many = kwargs.get("many", False)
        if "many" in kwargs.keys():
            del kwargs["many"]
        for key in kwargs.keys():
            if key not in cls._cols:
                raise AttributeError(f"Model {cls.__tablename__} don't have attribute {key}")

        # TODO: none value

        conditions = {}
        for k, v in list(kwargs.items()):
            conditions[k] = v
        if many:
            items = cls.query.filter_by(**conditions).all()
            return items

        item = cls.query.filter_by(**conditions).one_or_none()
        return item



import pathlib

from sqlalchemy.engine import Engine
from sqlalchemy import event

HERE = pathlib.Path(__file__).parent

version = open((HERE / "version"), "r").read().strip()


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


from . import api, task



from datetime import datetime

from paddlelabel.config import db
from paddlelabel.api.util import nncol

from .base import BaseModel


class User(BaseModel):
    __tablename__ = "user"
    __table_args__ = {"comment": "Stores all user info"}
    user_id = nncol(db.Integer, primary_key=True)
    uuid = nncol(db.String(), unique=True)
    username = nncol(db.String(), unique=True)
    email = nncol(db.String())
    password = nncol(db.String())
    role_id = nncol(db.Integer())

    _immutables = BaseModel._immutables + ["user_id", "uuid"]



import pathlib
from setuptools import setup, find_packages, Extension

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text(encoding="utf-8")


with open("requirements.txt", "r") as fin:
    REQUIRED_PACKAGES = fin.read()

setup(
    name="paddlelabel",
    version=open((HERE / "paddlelabel" / "version"), "r").read().strip(),
    description="Web Based Multi Purpose Annotation",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/PaddleCV-SIG/PP-Label",
    author="PaddleCV-SIG",
    author_email="me@linhan.email",
    license="Apache Software License",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=("test", "tool")),
    include_package_data=True,
    install_requires=REQUIRED_PACKAGES,
    entry_points={
        "console_scripts": [
            "paddlelabel=paddlelabel.__main__:run",
            "pdlabel=paddlelabel.__main__:run",
        ]
    },
)



import json
import time

from marshmallow import post_load, pre_load, pre_dump, post_dump, fields
from marshmallow.fields import Nested

from paddlelabel.api.model import Task
from paddlelabel.api.schema.util import str2sault
from .base import BaseSchema


class TaskSchema(BaseSchema):
    class Meta(BaseSchema.Meta):
        model = Task

    # project = Nested("ProjectSchema")
    annotations = fields.List(Nested("AnnotationSchema"), exclude=("task",))
    data_paths = fields.List(fields.String())
    # datas = fields.List(Nested("DataSchema"), exclude=("task",))

    # # TODO: confirm data['result'] dont cause trouble
    @pre_dump
    def output(self, task, **kwargs):
        paths = []
        for data in task.datas:
            paths.append(f"/datas/{data.data_id}/image?sault={str2sault(data.path)}")
        task.data_paths = paths
        return task



import json

from paddlelabel.config import db
from paddlelabel.api.util import nncol

# TODO: circular risk
from .base import BaseModel


class Project(BaseModel):
    __tablename__ = "project"
    __table_args__ = {"comment": "Stores information and settings for each project"}
    project_id = nncol(db.Integer, primary_key=True)
    name = nncol(db.String(length=256), unique=True)
    description = db.Column(db.String(length=256))
    task_category_id = db.Column(db.Integer(), db.ForeignKey("taskCategory.task_category_id"))
    task_category = db.relationship("TaskCategory")
    data_dir = nncol(db.String(length=256), unique=True)
    label_dir = db.Column(db.String(length=256), unique=True)
    labels = db.relationship(
        "Label",
        lazy="selectin",
        cascade="all, delete-orphan",
    )
    tasks = db.relationship("Task", lazy="noload", cascade="all, delete-orphan")
    label_format = db.Column(db.String())
    other_settings = db.Column(db.String())

    _immutables = BaseModel._immutables + ["project_id", "task_category_id"]

    def _get_other_settings(self):
        if self.other_settings is None:
            return {}
        return json.loads(self.other_settings)



import argparse
import os.path as osp
import os

parser = argparse.ArgumentParser(description="Remove first value from labels.txt")
parser.add_argument(
    "-l", "--labels_path", type=str, required=True, help="Path to labels.txt. Use absolute path to be safe."
)
parser.add_argument(
    "-o",
    "--output_path",
    type=str,
    required=True,
    help="Path to output processed labels.txt. Use absolute path to be safe.",
)
parser.add_argument(
    "-id",
    "--input_delimiter",
    default=" ",
    type=str,
    help='Delimiter for input labels.txt, warped with "". For example " " for space or "|" for |. Defaults to space',
)

parser.add_argument(
    "-od",
    "--output_delimiter",
    default=" ",
    type=str,
    help="Delimiter for output labels.txt, same format as input_delimiter. Defaults to space",
)

args = parser.parse_args()

if not osp.exists(args.labels_path):
    raise RuntimeError(f"Didn't find labels.txt at {args.labels_path}.")

lines = open(args.labels_path, "r").readlines()
lines = [l.strip() for l in lines if len(l.strip()) != 0]
lines = [l.split(args.input_delimiter) for l in lines]
lines = [l[1:] for l in lines]
lines = [args.output_delimiter.join(l) for l in lines]

os.makedirs(osp.dirname(args.output_path), exist_ok=True)
with open(args.output_path, "w") as f:
    for line in lines:
        print(line, file=f)



from paddlelabel.config import db
from paddlelabel.api.util import nncol
from .base import BaseModel


class TagTask(BaseModel):
    __tablename__ = "tagTask"
    __table_args__ = {"comment": "Tag and task intersect"}
    tag_task_id = nncol(db.Integer(), primary_key=True)
    project_id = nncol(db.Integer(), db.ForeignKey("project.project_id"))
    tag_id = nncol(db.Integer(), db.ForeignKey("tag.tag_id"))
    task_id = nncol(db.Integer(), db.ForeignKey("task.task_id"))
    tag = db.relationship("Tag")

    _immutables = BaseModel._immutables + [
        "tag_task_id",
        "project_id",
        "tag_id",
        "task_id",
    ]



from .project import Project
from .label import Label
from .task import Task
from .data import Data
from .annotation import Annotation
from .user import User
from .setting import TaskCategory
from .tag import Tag
from .tag_task import TagTask



import os.path as osp
import json
from copy import deepcopy

from pycocotoolse.coco import COCO
import cv2

from paddlelabel.api import Task, Annotation, Label
from paddlelabel.task.util import create_dir, listdir, copy, image_extensions
from paddlelabel.task.base import BaseTask
from paddlelabel.config import db
from paddlelabel.api.util import abort

# TODO: move to io
def parse_voc_label(label_path):
    from xml.dom import minidom

    def data(elements):
        return elements[0].firstChild.data

    file = minidom.parse(label_path)

    # 1. parse img info
    img = {}

    # 1.1 file path
    folder = file.getElementsByTagName("folder")
    if len(folder) == 0:
        folder = "JPEGImages"
    else:
        folder = data(folder)
    filename = file.getElementsByTagName("filename")
    if len(filename) == 0:
        abort(detail=f"Missing required field filename in annotation file {label_path}", status=404)
    filename = data(filename)
    path = osp.join(folder, filename)

    # 1.2 size
    size = file.getElementsByTagName("size")[0]
    size = [data(size.getElementsByTagName(n)) for n in ["width", "height"]]
    width, height = [int(t) for t in size]
    size = [str(t) for t in [1, width, height]]

    img["size"] = ",".join(size)
    img["path"] = path

    # 2. parse annotations
    objects = file.getElementsByTagName("object")
    anns = []
    frontend_id = 1
    for object in objects:
        ann = {}
        ann["label_name"] = data(object.getElementsByTagName("name"))
        bndbox = object.getElementsByTagName("bndbox")[0]
        # ann["result"] = {}
        # ann["result"]["xmin"] = data(bndbox.getElementsByTagName("xmin"))
        # ann["result"]["xmax"] = data(bndbox.getElementsByTagName("xmax"))
        # ann["result"]["ymin"] = data(bndbox.getElementsByTagName("ymin"))
        # ann["result"]["ymax"] = data(bndbox.getElementsByTagName("ymax"))
        # ann["result"] = json.dumps(ann["result"])
        names = ["xmin", "ymin", "xmax", "ymax"]
        r = [data(bndbox.getElementsByTagName(n)) for n in names]
        r = [int(t) for t in r]
        r[0] -= width / 2
        r[1] -= height / 2
        r[2] -= width / 2
        r[3] -= height / 2
        r = [str(t) for t in r]
        ann["result"] = ",".join(r)
        ann["type"] = "rectangle"
        ann["frontend_id"] = frontend_id
        anns.append(ann)
        frontend_id += 1
    return img, anns


def create_voc_label(filepath, width, height, annotations):
    object_labels = ""
    width = int(width)
    height = int(height)
    for ann in annotations:
        r = [float(t) for t in ann.result.split(",")]
        r[0] += width / 2
        r[1] += height / 2
        r[2] += width / 2
        r[3] += height / 2
        r = [int(t) for t in r]

        object_labels += f"""
            <object>
            <name>{ann.label.name}</name>
            <bndbox>
            <xmin>{r[0]}</xmin>
            <ymin>{r[1]}</ymin>
            <xmax>{r[2]}</xmax>
            <ymax>{r[3]}</ymax>
            </bndbox>
            </object>
        """

    folder = osp.dirname(filepath)
    filename = osp.basename(filepath)
    voc_label = f"""
        <?xml version='1.0' encoding='UTF-8'?>
        <annotation>
        <folder>{folder}</folder>
        <filename>{filename}</filename>
        <object_num>{len(annotations)}</object_num>
        <size>
            <width>{width}</width>
            <height>{height}</height>
        </size>
        {object_labels}
        </annotation>
    """

    # TODO: beautify export xml
    # from xml.dom import minidom
    # return minidom.parseString(voc_label.strip()).toprettyxml(indent="    ", newl="")
    return voc_label.strip()


# TODO: change data_dir to dataset_path


class Detection(BaseTask):
    def __init__(self, project, data_dir=None):
        super(Detection, self).__init__(project, data_dir=data_dir)
        self.importers = {"coco": self.coco_importer, "voc": self.voc_importer}
        self.exporters = {
            "coco": self.coco_exporter,
            "voc": self.voc_exporter,
        }
        self.default_importer = self.voc_importer
        self.default_exporter = self.voc_exporter  # default to voc

    def coco_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        """
        images should be located at data_dir / file_name in coco annotation
        """

        # 1. set params
        project = self.project
        if data_dir is None:
            data_dir = project.data_dir
        label_file_paths = ["train.json", "val.json", "test.json"]
        label_file_paths = [osp.join(data_dir, f) for f in label_file_paths]

        self.create_warning(data_dir)

        def _coco_importer(data_paths, label_file_path, set=0):
            coco = COCO(label_file_path)
            info = coco.dataset.get("info", {})
            licenses = coco.dataset.get("licenses", [])

            # 1. create all labels
            self.create_coco_labels(coco.cats.values())

            ann_by_task = {}
            # 2. get image full path and size
            for idx, img in coco.imgs.items():
                file_name = img["file_name"]
                full_path = filter(
                    lambda p: osp.normpath(p)[-len(osp.normpath(file_name)) :] == osp.normpath(file_name), data_paths
                )
                full_path = list(full_path)
                if len(full_path) != 1:
                    raise RuntimeError(
                        f"{'No' if len(full_path) == 0 else 'Multiple'} image(s) with path ending with {file_name} found under {data_dir}"
                    )
                full_path = full_path[0]
                data_paths.remove(full_path)
                coco.imgs[idx]["full_path"] = full_path
                s = [img.get("width", 0), img.get("height", 0)]
                if s == [0, 0]:
                    s = cv2.imread(full_path).shape[:2][::-1]
                s = [str(t) for t in s]
                coco.imgs[idx]["size"] = ",".join(s)
                ann_by_task[img["id"]] = []

            # 3. get ann by image
            for ann_id in coco.getAnnIds():
                ann = coco.anns[ann_id]
                label_name = coco.cats[ann["category_id"]]["name"]
                # result = {}
                # result["xmin"] = ann["bbox"][0]
                # result["ymin"] = ann["bbox"][1]
                # result["xmax"] = result["xmin"] + ann["bbox"][2]
                # result["ymax"] = result["ymin"] + ann["bbox"][3]
                # image center as origin, right x down y
                res = ann["bbox"]
                width, height = (
                    coco.imgs[ann["image_id"]].get("width", None),
                    coco.imgs[ann["image_id"]].get("height", None),
                )
                res[2] += res[0]
                res[3] += res[1]
                res[0] -= width / 2
                res[1] -= height / 2
                res[2] -= width / 2
                res[3] -= height / 2

                res = [str(r) for r in res]
                res = ",".join(res)
                # curr_anns = ann_by_task.get(ann["image_id"], [])
                ann_by_task[ann["image_id"]].append(
                    {
                        "label_name": label_name,
                        "result": res,
                        "type": "rectangle",
                        "frontend_id": len(ann_by_task[ann["image_id"]]) + 1,
                    }
                )

            # 4. add tasks
            for img_id, annotations in list(ann_by_task.items()):
                data_path = coco.imgs[img_id]["full_path"]
                size = "1," + coco.imgs[img_id]["size"]
                self.add_task([{"path": data_path, "size": size}], [annotations], split=set)
            return data_paths, json.dumps({"info": info, "licenses": licenses})

        # 2. find all images under data_dir
        data_paths = listdir(data_dir, filters=filters)
        coco_others = {}
        for split_idx, label_file_path in enumerate(label_file_paths):
            if osp.exists(label_file_path):
                data_paths, others = _coco_importer(data_paths, label_file_path, split_idx)
                coco_others[split_idx] = others
        other_settings = project._get_other_settings()
        other_settings["coco_others"] = coco_others
        project.other_settings = json.dumps(other_settings)

        # 3. add tasks without label
        for data_path in data_paths:
            img = cv2.imread(osp.join(data_dir, data_path))
            s = img.shape
            size = [1, s[1], s[0], s[2]]
            size = [str(s) for s in size]
            size = ",".join(size)
            self.add_task([{"path": data_path, "size": size}])

        db.session.commit()

    def voc_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        # 1. set params
        project = self.project
        base_dir = data_dir
        allow_missing_image = data_dir is not None

        if base_dir is None:
            base_dir = project.data_dir
        self.create_warning(base_dir)

        # 2. get all data and label
        data_paths = set(p for p in listdir(base_dir, filters=filters))
        label_paths = [p for p in listdir(base_dir, filters={"exclude_prefix": ["."], "include_postfix": [".xml"]})]

        # print("data_paths", data_paths)
        # print("label_paths", label_paths)

        for label_path in label_paths:
            data, labels = parse_voc_label(osp.join(base_dir, label_path))
            if not osp.exists(osp.join(base_dir, data["path"])):
                if allow_missing_image:
                    continue
                else:
                    raise RuntimeError(f"Image specified in {label_path} not found.")
            self.add_task([data], [labels])
            data_paths.remove(data["path"])

        for data_path in data_paths:
            img = cv2.imread(osp.join(base_dir, data_path))
            s = img.shape
            size = [1, s[1], s[0], s[2]]
            size = [str(s) for s in size]
            size = ",".join(size)
            self.add_task([{"path": data_path, "size": size}])

        db.session.commit()

    def coco_exporter(self, export_dir):
        # 1. set params
        project = self.project

        # 2. create coco with all tasks
        coco = COCO()
        # 2.1 add categories
        labels = Label._get(project_id=project.project_id, many=True)
        for label in labels:
            if label.super_category_id is None:
                super_category_name = "none"
            else:
                super_category_name = self.label_id2name(label.super_category_id)
            coco.addCategory(label.id, label.name, label.color, super_category_name)

        # 2.2 add images
        split = [set(), set(), set()]
        tasks = Task._get(project_id=project.project_id, many=True)
        data_dir = osp.join(export_dir, "image")
        create_dir(data_dir)
        for task in tasks:
            data = task.datas[0]
            size = data.size.split(",")
            export_path = osp.join("image", osp.basename(data.path))
            coco.addImage(export_path, int(size[1]), int(size[2]), data.data_id)
            copy(osp.join(project.data_dir, data.path), data_dir)
            split[task.set].add(data.data_id)

        # 2.3 add annotations
        annotations = Annotation._get(project_id=project.project_id, many=True)
        for ann in annotations:
            r = ann.result.split(",")
            r = [float(t) for t in r]
            # print(coco.imgs[ann.data_id])
            width, height = (
                coco.imgs[ann.data_id]["width"],
                coco.imgs[ann.data_id]["height"],
            )
            width = int(width)
            height = int(height)
            bb = [
                r[0] + width / 2,
                r[1] + height / 2,
                r[2] + width / 2,
                r[3] + height / 2,
            ]
            bb[2] -= bb[0]
            bb[3] -= bb[1]
            coco.addAnnotation(
                ann.data_id,
                ann.label.id,
                segmentation=[],
                id=ann.annotation_id,
                bbox=bb,
            )

        # 3. write coco json
        coco_others = project._get_other_settings().get("coco_others", {})
        for split_idx, fname in enumerate(["train.json", "val.json", "test.json"]):
            outcoco = deepcopy(coco)
            outcoco.dataset["images"] = [img for img in coco.dataset["images"] if img["id"] in split[split_idx]]
            outcoco.dataset["annotations"] = [
                ann for ann in coco.dataset["annotations"] if ann["image_id"] in split[split_idx]
            ]

            coco_others_split = coco_others.get(str(split_idx), "{}")
            coco_others_split = json.loads(coco_others_split)

            outcoco.dataset["info"] = coco_others_split.get("info", "")
            outcoco.dataset["licenses"] = coco_others_split.get("licenses", [])

            with open(osp.join(export_dir, fname), "w") as outf:
                print(json.dumps(outcoco.dataset), file=outf)

    def voc_exporter(self, export_dir):
        # 1. set params
        project = self.project

        export_data_dir = osp.join(export_dir, "JPEGImages")
        export_label_dir = osp.join(export_dir, "Annotations")
        create_dir(export_data_dir)
        create_dir(export_label_dir)

        self.export_labels(export_dir)

        tasks = Task._get(project_id=project.project_id, many=True)
        export_paths = []

        for task in tasks:
            data = task.datas[0]
            data_path = osp.join(project.data_dir, data.path)
            export_path = osp.join("JPEGImages", osp.basename(data.path))

            copy(data_path, export_data_dir)
            id = osp.basename(data_path).split(".")[0]
            width, height = data.size.split(",")[1:3]
            with open(osp.join(export_label_dir, f"{id}.xml"), "w") as f:
                print(
                    create_voc_label(export_path, width, height, task.annotations),
                    file=f,
                )
            export_paths.append([export_path])

        self.export_split(export_dir, tasks, export_paths, with_labels=False, annotation_ext=".xml")



import os.path as osp
import json
from copy import deepcopy

import tifffile as tif
import numpy as np
import cv2
from pycocotoolse.coco import COCO

from paddlelabel.task.util import create_dir, listdir, image_extensions
from paddlelabel.task.base import BaseTask
from paddlelabel.config import db
from paddlelabel.task.util.color import hex_to_rgb
from paddlelabel.task.util import copy
from paddlelabel.api.model import Task, Label, Annotation
from paddlelabel.api.util import abort
from paddlelabel.api.rpc.seg import polygon2points

# debug
# import matplotlib
# matplotlib.use("TkAgg")
# import matplotlib.pyplot as plt


def draw_mask(data, type="pesudo"):
    height, width = map(int, data.size.split(",")[1:3])
    if type == "pesudo":
        mask = np.zeros((height, width, 3))
    else:
        mask = np.zeros((height, width))

    for ann in data.annotations:
        if ann.type != "brush":
            continue

        # TODO: remove
        if ann.result[:2] == "[[":
            continue

        label_id = ann.label.id
        result = ann.result.strip().split(",")
        try:
            result = [int(float(p)) for p in result]
        except:
            print(result, "to float error, plz open an issue for this")
        if ann.type == "brush":
            points = result[2:]
            line_width = result[0]
            if result[1] == 0:
                frontend_id = 0
                label_id = 0
            prev_w, prev_h = points[0:2]
        else:
            for idx in range(0, len(result), 2):
                result[idx] = int(result[idx] + width / 2)
                result[idx + 1] = int(result[idx + 1] + height / 2)
            points = polygon2points(result)
            points = np.array(points).reshape((-1))
            line_width = 1
            prev_w, prev_h = points[0:2]
        try:
            for idx in range(2, len(points), 2):
                w, h = points[idx : idx + 2]
                if type == "pesudo":
                    color = hex_to_rgb(ann.label.color)[::-1]
                else:
                    color = int(label_id)
                if line_width == 0:
                    line_width = 1
                cv2.line(mask, (prev_w, prev_h), (w, h), color, line_width)
                prev_w, prev_h = w, h
        except Exception as e:
            abort(detail=e.msg, status=500, title="cv2 error")
    return mask


def parse_semantic_mask(annotation_path, labels):
    ann = cv2.imread(annotation_path, cv2.IMREAD_UNCHANGED)
    frontend_id = 1
    anns = []
    # TODO: len(ann.shape == 3) and ann.shape[-1] == 1 necessary?
    # print(labels)
    if len(ann.shape) == 3:
        ann = cv2.cvtColor(ann, cv2.COLOR_BGR2RGB)
        ann_gray = np.zeros(ann.shape[:2], dtype="uint8")
        for label in labels:
            color = hex_to_rgb(label.color)
            label_mask = np.all(ann == color, axis=2)
            ann_gray[label_mask == 1] = label.id
            ann[label_mask == 1] = 0

        if ann.sum() != 0:
            ann = ann.reshape((-1, ann.shape[-1]))
            abort(
                f"Mask {annotation_path} contains unspecified labels {np.unique(ann, axis=0)[1:].tolist()} . Maybe you didn't include a background class in the first line of labels.txt or didn't specify label color?",
                404,
            )

        ann = ann_gray

    for label in labels:
        label_mask = deepcopy(ann)
        label_mask[label_mask != label.id] = 0
        label_mask[label_mask != 0] = 255

        if label_mask.sum() == 0:
            continue

        ann[ann == label.id] = 0
        (cc_num, cc_mask, values, centroid) = cv2.connectedComponentsWithStats(label_mask, connectivity=8)
        for cc_id in range(1, cc_num):
            h, w = np.where(cc_mask == cc_id)
            result = ",".join([f"{w},{h}" for h, w in zip(h, w)])
            result = f"{1},{frontend_id}," + result
            anns.append(
                {
                    "label_name": label.name,
                    "result": result,
                    "type": "brush",
                    "frontend_id": label.id,
                }
            )
            frontend_id += 1

    if ann.sum() != 0:
        msg = f"Mask {annotation_path} contains unspecified labels {np.unique(ann)[1:].tolist()} . Maybe you didn't include a background class in the first line of labels.txt or didn't specify label id?"
        print(msg)
        abort(msg, 404)

    s = [1] + list(ann.shape)
    s = [str(s) for s in s]
    size = ",".join(s)
    return size, anns


def parse_instance_mask(annotation_path, labels):
    mask = tif.imread(annotation_path)
    instance_mask = mask[0]
    label_mask = mask[1]
    anns = []

    for label in labels:
        instance_part = instance_mask[label_mask == label.id]
        instance_ids = np.unique(instance_part)
        for instance_id in instance_ids:
            h, w = np.where(instance_mask == instance_id)
            result = ",".join([f"{w},{h}" for w, h in zip(w, h)])
            result = f"{1},{instance_id}," + result
            anns.append(
                {
                    "label_name": label.name,
                    "result": result,
                    "type": "brush",
                    "frontend_id": str(instance_id),
                }
            )
    s = [1] + list(instance_mask.shape)
    s = [str(s) for s in s]
    size = ",".join(s)
    return size, anns


class InstanceSegmentation(BaseTask):
    def __init__(self, project, data_dir=None):
        super().__init__(project, skip_label_import=True, data_dir=data_dir)
        self.importers = {
            "mask": self.mask_importer,
            "polygon": self.coco_importer,
        }
        self.exporters = {
            "mask": self.mask_exporter,
            "polygon": self.coco_exporter,
        }
        self.default_importer = self.coco_importer
        self.default_exporter = self.coco_exporter

    def mask_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        # 1. set params
        project = self.project
        if data_dir is None:
            data_dir = project.data_dir

        background_line = self.import_labels(ignore_first=True)
        other_settings = project._get_other_settings()
        other_settings["background_line"] = background_line
        project.other_settings = json.dumps(other_settings)

        ann_dict = {
            osp.basename(p).split(".")[0]: p
            for p in listdir(data_dir, {"exclude_prefix": ["."], "include_postfix": [".tiff", ".tif"]})
        }

        # 2. import records
        for data_path in listdir(data_dir, filters):
            id = osp.basename(data_path).split(".")[0]
            data_path = osp.join(data_dir, data_path)
            if id in ann_dict.keys():
                ann_path = osp.join(data_dir, ann_dict[id])
                size, anns = parse_instance_mask(ann_path, project.labels)
            else:
                anns = []
                img = cv2.imread(data_path)
                s = [1] + list(img.shape)
                size = ",".join([str(s) for s in s])

            self.add_task([{"path": data_path, "size": size}], [anns])
        db.session.commit()

    def mask_exporter(self, export_dir):
        # 1. set params
        project = self.project

        # 2. create export destinations
        export_data_dir = osp.join(export_dir, "JPEGImages")
        export_label_dir = osp.join(export_dir, "Annotations")
        create_dir(export_data_dir)
        create_dir(export_label_dir)

        tasks = Task._get(project_id=project.project_id, many=True)
        export_data_paths = []
        export_label_paths = []

        for task in tasks:
            data = task.datas[0]
            data_path = osp.join(project.data_dir, data.path)
            export_data_path = osp.join("JPEGImages", osp.basename(data.path))
            # TODO: strip ext
            export_label_path = osp.join(export_label_dir, osp.basename(data_path).split(".")[0] + ".tiff")

            copy(data_path, export_data_dir)
            height, width = map(int, data.size.split(",")[1:3])

            instance_mask = np.zeros((height, width), dtype="uint8")
            label_mask = np.zeros((height, width), dtype="uint8")
            for ann in task.annotations:

                # TODO: skip eiseg result, remove this
                if ann.result[:2] == "[[":
                    continue

                label_id = ann.label.id
                frontend_id = ann.frontend_id
                result = ann.result.split(",")
                try:
                    result = [int(float(p)) for p in result]
                except:
                    print(result, "to float error, plz open an issue for this")
                if ann.type == "brush":
                    points = result[2:]
                    line_width = result[0]
                    if result[1] == 0:
                        frontend_id = 0
                        label_id = 0
                    prev_w, prev_h = points[0:2]
                else:
                    for idx in range(0, len(result), 2):
                        result[idx] = int(result[idx] + width / 2)
                        result[idx + 1] = int(result[idx + 1] + height / 2)
                    points = polygon2points(result)
                    points = np.array(points).reshape((-1))
                    line_width = 1
                    prev_w, prev_h = points[0:2]
                try:
                    for idx in range(2, len(points), 2):
                        w, h = points[idx : idx + 2]
                        if line_width == 0:
                            line_width = 1

                        cv2.line(
                            label_mask,
                            (prev_w, prev_h),
                            (w, h),
                            int(label_id),
                            line_width,
                        )
                        cv2.line(
                            instance_mask,
                            (prev_w, prev_h),
                            (w, h),
                            int(frontend_id),
                            line_width,
                        )
                        prev_w, prev_h = w, h
                except Exception as e:
                    abort(detail=e.msg, status=500, title="cv2 error")

            mask = np.stack([instance_mask, label_mask], axis=0)
            tif.imwrite(export_label_path, mask, compression="zlib")

            export_data_paths.append([export_data_path])
            export_label_paths.append([export_label_path])

        self.export_split(
            export_dir,
            tasks,
            export_data_paths,
            with_labels=False,
            annotation_ext=".tiff",
        )
        background_line = project._get_other_settings().get("background_line")
        if background_line is None or len(background_line) == 0:
            background_line = "background"
        self.export_labels(export_dir, background_line, with_id=True)

    def coco_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):
        # 1. set params
        project = self.project
        if data_dir is None:
            data_dir = project.data_dir
        label_file_paths = ["train.json", "val.json", "test.json"]
        label_file_paths = [osp.join(data_dir, f) for f in label_file_paths]

        self.create_warning(data_dir)

        def _coco_importer(data_paths, label_file_path, set=0):
            coco = COCO(label_file_path)
            info = coco.dataset.get("info", {})
            licenses = coco.dataset.get("licenses", [])

            # 1. create all labels
            self.create_coco_labels(coco.cats.values())

            print("=====", data_paths)
            ann_by_task = {}
            # 2. get image full path and size
            for idx, img in coco.imgs.items():
                file_name = img["file_name"]
                full_path = filter(
                    lambda p: osp.normpath(p)[-len(osp.normpath(file_name)) :] == osp.normpath(file_name), data_paths
                )
                full_path = list(full_path)
                if len(full_path) != 1:
                    abort(
                        detail=f"{'No' if len(full_path) == 0 else 'Multiple'} image(s) with path ending with {file_name} found under {data_dir}",
                        status=404,
                    )

                full_path = full_path[0]
                data_paths.remove(full_path)
                coco.imgs[idx]["full_path"] = full_path
                s = [img.get("height", 0), img.get("width", 0)]
                if s == [0, 0]:
                    s = cv2.imread(full_path).shape[:2]
                s = [str(t) for t in s]
                coco.imgs[idx]["size"] = ",".join(s)
                ann_by_task[img["id"]] = []

            # 3. get ann by image
            for ann_id in coco.getAnnIds():
                ann = coco.anns[ann_id]
                if coco.imgs.get(ann["image_id"]) is None:
                    print(f"No image with id {ann['image_id']} found, skipping this annotation.")
                    continue

                label_name = coco.cats[ann["category_id"]]["name"]
                res = ann["segmentation"][0]
                width, height = (
                    coco.imgs[ann["image_id"]].get("width", None),
                    coco.imgs[ann["image_id"]].get("height", None),
                )
                for idx in range(0, len(res), 2):
                    res[idx] -= width / 2
                    res[idx + 1] -= height / 2

                res = [str(r) for r in res]
                res = ",".join(res)
                ann_by_task[ann["image_id"]].append(
                    {
                        "label_name": label_name,
                        "result": res,
                        "type": "polygon",
                        "frontend_id": len(ann_by_task[ann["image_id"]]) + 1,
                    }
                )

            # 4. add tasks
            for img_id, annotations in list(ann_by_task.items()):
                data_path = coco.imgs[img_id]["full_path"]
                size = "1," + coco.imgs[img_id]["size"]
                self.add_task([{"path": data_path, "size": size}], [annotations], split=set)
            return data_paths, json.dumps({"info": info, "licenses": licenses})

        # 2. find all images under data_dir
        data_paths = listdir(data_dir, filters=filters)
        coco_others = {}
        for split_idx, label_file_path in enumerate(label_file_paths):
            if osp.exists(label_file_path):
                data_paths, others = _coco_importer(data_paths, label_file_path, split_idx)
                coco_others[split_idx] = others
        other_settings = project._get_other_settings()
        other_settings["coco_others"] = coco_others
        project.other_settings = json.dumps(other_settings)

        # 3. add tasks without label
        for data_path in data_paths:
            img = cv2.imread(osp.join(data_dir, data_path))
            s = img.shape
            size = [1, s[1], s[0], s[2]]
            size = [str(s) for s in size]
            size = ",".join(size)
            self.add_task([{"path": data_path, "size": size}])

        db.session.commit()

    def coco_exporter(self, export_dir):
        # 1. set params
        project = self.project

        # 2. create coco with all tasks
        coco = COCO()
        # 2.1 add categories
        labels = Label._get(project_id=project.project_id, many=True)
        for label in labels:
            if label.super_category_id is None:
                super_category_name = "none"
            else:
                super_category_name = self.label_id2name(label.super_category_id)
            coco.addCategory(label.id, label.name, label.color, super_category_name)

        # 2.2 add images
        split = [set(), set(), set()]
        tasks = Task._get(project_id=project.project_id, many=True)
        data_dir = osp.join(export_dir, "image")
        create_dir(data_dir)
        for task in tasks:
            data = task.datas[0]
            size = data.size.split(",")
            export_path = osp.join("image", osp.basename(data.path))
            coco.addImage(export_path, int(size[1]), int(size[2]), data.data_id)
            copy(osp.join(project.data_dir, data.path), data_dir)
            split[task.set].add(data.data_id)

        # 2.3 add annotations
        annotations = Annotation._get(project_id=project.project_id, many=True)
        for ann in annotations:
            if ann.type != "polygon":
                continue

            r = ann.result.split(",")
            r = [float(t) for t in r]
            width, height = (
                coco.imgs[ann.data_id]["width"],
                coco.imgs[ann.data_id]["height"],
            )
            width = int(width)
            height = int(height)
            for idx in range(0, len(r), 2):
                r[idx] += width / 2
                r[idx + 1] += height / 2

            coco.addAnnotation(
                ann.data_id,
                ann.label.id,
                segmentation=r,
                id=ann.annotation_id,
            )

        # 3. write coco json
        coco_others = project._get_other_settings().get("coco_others", {})
        for split_idx, fname in enumerate(["train.json", "val.json", "test.json"]):
            outcoco = deepcopy(coco)
            outcoco.dataset["images"] = [img for img in coco.dataset["images"] if img["id"] in split[split_idx]]
            outcoco.dataset["annotations"] = [
                ann for ann in coco.dataset["annotations"] if ann["image_id"] in split[split_idx]
            ]

            coco_others_split = coco_others.get(str(split_idx), "{}")
            coco_others_split = json.loads(coco_others_split)

            outcoco.dataset["info"] = coco_others_split.get("info", "")
            outcoco.dataset["licenses"] = coco_others_split.get("licenses", [])

            with open(osp.join(export_dir, fname), "w") as outf:
                print(json.dumps(outcoco.dataset), file=outf)


class SemanticSegmentation(InstanceSegmentation):
    def __init__(self, project, data_dir=None):
        super().__init__(project, data_dir=data_dir)
        self.importers = {
            "mask": self.mask_importer,
            "polygon": self.coco_importer,
        }
        self.exporters = {
            "mask": self.mask_exporter,
            "polygon": self.coco_exporter,
        }
        self.default_importer = self.mask_importer
        self.default_exporter = self.mask_exporter

    def mask_importer(
        self,
        data_dir=None,
        filters={"exclude_prefix": ["."], "include_postfix": image_extensions},
    ):

        # 1. set params
        project = self.project
        if data_dir is None:
            base_dir = project.data_dir
        else:
            base_dir = data_dir
        data_dir = osp.join(base_dir, "JPEGImages")
        ann_dir = osp.join(base_dir, "Annotations")

        background_line = self.import_labels(ignore_first=True)
        other_settings = project._get_other_settings()
        other_settings["background_line"] = background_line
        project.other_settings = json.dumps(other_settings)

        ann_dict = {osp.basename(p).split(".")[0]: p for p in listdir(ann_dir, filters)}

        # 2. import records
        for data_path in listdir(data_dir, filters):
            id = osp.basename(data_path).split(".")[0]
            data_path = osp.join(data_dir, data_path)
            if id in ann_dict.keys():
                ann_path = osp.join(ann_dir, ann_dict[id])
                size, anns = parse_semantic_mask(ann_path, project.labels)
            else:
                anns = []
                img = cv2.imread(data_path)
                s = [1] + list(img.shape)
                size = ",".join([str(s) for s in s])

            self.add_task([{"path": data_path, "size": size}], [anns])
        db.session.commit()

    def mask_exporter(self, export_dir: str):
        """Export semantic segmentation dataset in mask format

        Args:
            export_dir (str): The folder to export to.
            type (str, optional): Mask type, "gray" or "pesudo". Defaults to "gray".
        """

        # 1. set params
        project = self.project
        other_settings = project._get_other_settings()
        print(other_settings)
        type = other_settings.get("segMaskType", "grayscale")

        export_data_dir = osp.join(export_dir, "JPEGImages")
        export_label_dir = osp.join(export_dir, "Annotations")
        create_dir(export_data_dir)
        create_dir(export_label_dir)

        tasks = Task._get(project_id=project.project_id, many=True)
        export_data_paths = []
        export_label_paths = []

        for task in tasks:
            data = task.datas[0]
            data_path = osp.join(project.data_dir, data.path)

            export_data_path = osp.join("JPEGImages", osp.basename(data.path))

            # TODO: strip ext
            export_label_path = osp.join(export_label_dir, osp.basename(data_path).split(".")[0] + ".png")

            copy(data_path, export_data_dir)

            mask = draw_mask(data, type=type)
            cv2.imwrite(export_label_path, mask)

            export_data_paths.append([export_data_path])
            export_label_paths.append([export_label_path])

        self.export_split(
            export_dir,
            tasks,
            export_data_paths,
            with_labels=False,
            annotation_ext=".png",
        )
        bg = project._get_other_settings().get("background_line", "background")
        self.export_labels(export_dir, bg)



import os.path as osp
import argparse

from tqdm import tqdm

from util.file import listdir, create_dir, copy

parser = argparse.ArgumentParser(
    description="Move single class classification images to class name folders based on xx_list.txt and labels.txt\n Note: only support consecutive label id"
)
parser.add_argument(
    "-i",
    "--dataset_path",
    type=str,
    required=True,
    help="Dataset path, should have labels.txt, and at least one of train_list.txt, val_list.txt and test_list.txt",
)
parser.add_argument("-o", "--output_path", type=str, required=True, help="The path to export dataset to")
parser.add_argument(
    "-id",
    "--input_delimiter",
    default=" ",
    type=str,
    help='Delimiter for input labels.txt, warped with "". For example " " for space or "|" for |. Defaults to space',
)

parser.add_argument(
    "-od",
    "--output_delimiter",
    default=" ",
    type=str,
    help="Delimiter for output labels.txt, same format as input_delimiter. Defaults to space",
)

args = parser.parse_args()

# 1. ensure essential files exist
labes_txt_path = osp.join(args.dataset_path, "labels.txt")
assert osp.exists(labes_txt_path), f"labels.txt not found under {args.dataset_path}"

xx_list_found = 0
list_paths = [osp.join(args.dataset_path, n) for n in ("train_list.txt", "val_list.txt", "test_list.txt")]
for list_path in list_paths:
    if osp.exists(list_path):
        xx_list_found += 1
assert xx_list_found != 0, f"No xx_list.txt found under {args.dataset_path}"

# 2. read info
all_files = set(listdir(args.dataset_path))

labels = open(labes_txt_path, "r").readlines()
labels = [l.strip() for l in labels]
labels = [l for l in labels if len(l) != 0]

idx2label = {}
for idx, name in enumerate(labels):
    idx2label[idx] = name

list_lines = []
for list_path in list_paths:
    if osp.exists(list_path):
        list_lines += open(list_path, "r").readlines()
list_lines = [l.strip() for l in list_lines]
list_lines = [l for l in list_lines if len(l) != 0]
list_info = [l.split(args.input_delimiter) for l in list_lines]

# 3. create output dirs and move files
for label_name in idx2label.values():
    create_dir(osp.join(args.output_path, label_name))

for img_path, label_id in tqdm(list_info):
    label_name = idx2label[int(label_id)]
    copy(osp.join(args.dataset_path, img_path), osp.join(args.output_path, label_name))
    all_files.remove(img_path)

# move all remaining files
for file in all_files:
    copy(osp.join(args.dataset_path, file), osp.join(args.output_path, file))
